/*  Minibar - functions.c "generalized functions used by the app"
    Copyright (C) 2002-2005 Ray Vanderborght (depsypher@users.sf.net)
    This file is part of Minibar.

    Minibar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Minibar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Minibar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "minibar.h"
#include "protos.h"

//===========================================================================
//	Globals...
//===========================================================================

extern DmOpenRef 	gDrinkDB;
extern char			*gsNone;

//===========================================================================
// Functions...
//===========================================================================


/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	AddDrinkRecord
// DESCRIPTION: Add a drink record to the database
// PARAMETERS:
//
Err AddDrinkRecord(UInt16 nIndex, DrinkPtr pDrink)
{
	void				*pRec;
	char 				zero = 0;
	UInt8				n;
	UInt8				*pDescriptor;
	IngredientNodePtr	pNode;
	IngredientType		ingredient;
	UInt32 				nOffset;
	UInt32 				nSize;
	UInt32				nLength;
	MemHandle 			hNewRec;
	Err					error = errNone;

	// create the new record
	nSize = DrinkPackedSize(pDrink);
	hNewRec = DmNewRecord(gDrinkDB, &nIndex, nSize);
	if(!hNewRec)
		return DmGetLastErr();
	pRec = MemHandleLock(hNewRec);

	// Write out the data	//

	nOffset = 0;
	DmWrite(pRec, nOffset, &pDrink->nIngredients, sizeof(UInt8));
	nOffset++;

	pNode = pDrink->pIngredientList;
	for(n = 0; n < pDrink->nIngredients; n++)
	{
		DmWrite(pRec, nOffset, &pNode->ingredient, sizeof(IngredientType));
		nOffset += sizeof(IngredientType);
		if(pNode->pNext)
			pNode = pNode->pNext;
	}

	DmWrite(pRec, nOffset, &pDrink->nDescriptors, sizeof(UInt8));
	nOffset++;

	pDescriptor = pDrink->pDescriptors;
	for(n = 0; n < pDrink->nDescriptors; n++)
	{
		DmWrite(pRec, nOffset, pDescriptor, sizeof(UInt8));
		nOffset++;
		pDescriptor++;
	}

	DmStrCopy(pRec, nOffset, pDrink->sName);
	nOffset += StrLen(pDrink->sName) + 1;

	DmStrCopy(pRec, nOffset, pDrink->sDirections);
//	nOffset += StrLen(pDrink->sDirections) + 1;

	MemHandleUnlock(hNewRec);

	return error;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	AddIngredientCleanUp
// DESCRIPTION: do what is needed when closing down the AddIngredient form
// PARAMETERS:
//
void AddIngredientCleanup(UInt8 nIngredients, UInt8 nDescriptors,
	char **psIngredients, char **psDescriptors, MemHandle hUnitsList,
	MemHandle hUnitsStrings)
{
	StrArrayFree(psIngredients, nIngredients);
	if(psIngredients)
	{
		ErrFatalDisplayIf(MemPtrFree(psIngredients), "delete failed");
		psIngredients = NULL;
	}

	StrArrayFree(psDescriptors, nDescriptors + 1);	// add one for "none" selection
	if(psDescriptors)
	{
		ErrFatalDisplayIf(MemPtrFree(psDescriptors), "delete failed");
		psDescriptors = NULL;
	}

	if(hUnitsList)
		ErrFatalDisplayIf(MemHandleFree(hUnitsList),
			"error freeing hUnitsList");

	if(hUnitsStrings)
	{
		ErrFatalDisplayIf(MemHandleUnlock(hUnitsStrings),
			"error unlocking hUnitStrings");
		DmReleaseResource(hUnitsStrings);
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	BuildIngredientString
// DESCRIPTION: get a string representation of a particular IngredientType
// PARAMETERS:
//
void BuildIngredientString(AppInfoBlockPtr pInfo, IngredientType ingredient,
	char *sNew)
{
	UInt8	n, *pTemp;
	char	s[SHORT_STRING];

	StrIToA(sNew, ingredient.nAmount / 4);
	if(sNew[0] == '0')	sNew[0] = NULL;
	n = StrLen(sNew);
	sNew[n] = GetFractionalLabel(ingredient.nAmount);
	if(sNew[n] == '-')
		sNew[n] = NULL;
	else
		sNew[n+1] = NULL;
	StrCat(sNew, " ");

	SysStringByIndex(UnitsStringTable, ingredient.nUnits, s, SHORT_STRING);
	StrCat(sNew, s);
	StrCat(sNew, " ");

	pTemp = pInfo->pIngredients;
	for(n = 0; n < ingredient.nName; n++)
		pTemp += StrLen(pTemp) + 1;
	StrCat(sNew, pTemp);

	if(ingredient.nDescriptors > 0)
	{
		StrCat(sNew, " (");
		pTemp = pInfo->pIngredientDescriptors;
		for(n = 0; n < ingredient.nDescriptor; n++)
			pTemp += StrLen(pTemp) + 1;
		StrCat(sNew, pTemp);
		StrCat(sNew, ")");
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	CopyDrinkDescriptors
// DESCRIPTION:	copies the drink descriptors to app info block, returns
//				the offset of the next byte to be written
// PARAMETERS:
//
void CopyDrinkDescriptors(UInt8 *pBlock, UInt16 *pnOffset, AppInfoBlockPtr info)
{
	UInt8 n;
	UInt8 *p;

	*pnOffset = 0;
	DmSet(pBlock, *pnOffset, 1, info->nDrinkDescriptors);
	*pnOffset = *pnOffset + 1;
	p = info->pDrinkDescriptors;
	for(n = 0; n < info->nDrinkDescriptors; n++)
	{
		DmStrCopy(pBlock, *pnOffset, p);
		*pnOffset = *pnOffset + StrLen(p) + 1;
		p += StrLen(p) + 1;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	CopyIngredientDescriptors
// DESCRIPTION:	copies the ingredient descriptors to appinfo block, returns
//				the offset of the next byte to be written
// PARAMETERS:
//
void CopyIngredientDescriptors(UInt8 *pBlock, UInt16 *pnOffset, AppInfoBlockPtr info)
{
	UInt8 n;
	UInt8 *p;

	DmSet(pBlock, *pnOffset, 1, info->nIngredientDescriptors);
	*pnOffset = *pnOffset + 1;
	p = info->pIngredientDescriptors;
	for(n = 0; n < info->nIngredientDescriptors; n++)
	{
		DmStrCopy(pBlock, *pnOffset, p);
		*pnOffset = *pnOffset + StrLen(p) + 1;
		p += StrLen(p) + 1;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	CopyIngredients
// DESCRIPTION:	copies the ingredients to appinfo block, returns
//				the offset of the next byte to be written
// PARAMETERS:
//
void CopyIngredients(UInt8 *pBlock, UInt16 *pnOffset, AppInfoBlockPtr info)
{
	UInt8 n;
	UInt8 *p;

	DmSet(pBlock, *pnOffset, 1, info->nIngredients);
	*pnOffset = *pnOffset + 1;
	p = info->pIngredients;
	for(n = 0; n < info->nIngredients; n++)
	{
		DmStrCopy(pBlock, *pnOffset, p);
		*pnOffset = *pnOffset + StrLen(p) + 1;
		p += StrLen(p) + 1;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	CompareDrinkRecords
// DESCRIPTION: compare two drinks for sorting alphabetically
// PARAMETERS:
//
Int16 CompareDrinkRecords(UInt8 *pRec1, UInt8 *pRec2, Int16 nOther,
	SortRecordInfoPtr pRec1SortInfo, SortRecordInfoPtr pRec2SortInfo,
	MemHandle hAppInfo)
{
	UInt8	*p1, *p2;
	UInt16	n, nItems;

	// find the offset of the drink name in the first record
	//
	p1 = pRec1;
	nItems = *(UInt8*)p1++;
	for(n = 0; n < nItems; n++)
		p1 += sizeof(IngredientType);

	nItems = *(UInt8*)p1++;
	if(nItems)
		p1++;

	// find the offset of the drink name in the second record
	//
	p2 = pRec2;
	nItems = *(UInt8*)p2++;

	for(n = 0; n < nItems; n++)
		p2 += sizeof(IngredientType);

	nItems = *(UInt8*)p2++;
	if(nItems)
		p2++;

	return StrCaselessCompare(p1, p2);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	DisplayInt
// DESCRIPTION: convenience function used for displaying an integer
// PARAMETERS:
//
void DisplayInt(UInt32 n)
{
	char s[32];
	StrIToA(s, n);
	FrmCustomAlert(Alert, "n=", s, NULL);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	DisplayString
// DESCRIPTION: convenience function for displaying a string
// PARAMETERS:
//
void DisplayString(char *s)
{
	FrmCustomAlert(Alert, s, NULL, NULL);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	DrawBoxedMessage
// DESCRIPTION: draws a message at the given coords with a box around it
// PARAMETERS:
//
void DrawBoxedMessage(UInt16 x, UInt16 y, char *sMessage)
{
	RectangleType rect1, rect2;
	UInt16 nLength;

	nLength = StrLen(sMessage);
	rect1.topLeft.x = x;
	rect1.topLeft.y = y;
	rect1.extent.x = 5 * nLength;
	rect1.extent.y = 25;
	rect2.topLeft.x = x + 2;
	rect2.topLeft.y = y + 2;
	rect2.extent.x = rect1.extent.x - 4;
	rect2.extent.y = rect1.extent.y - 4;
	WinDrawRectangle(&rect1, 3);
	WinEraseRectangle(&rect2, 3);
	WinDrawChars(sMessage, nLength, x + 5 * nLength / 2 - nLength * 2 + 2, y + 7);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	DrinkFree
// DESCRIPTION: free the memory that's allocated by UnpackDrinkRecord()
// PARAMETERS:
//
Err DrinkFree(DrinkPtr pDrink)
{
	IngredientNodePtr pNode, pTempNode;
	Err error = errNone;

	if(!pDrink)
		return errNone;

	if(pDrink->pIngredientList)
	{
		pNode = pDrink->pIngredientList;
		while(pNode)
		{
			pTempNode = pNode->pNext;
			ErrFatalDisplayIf(MemPtrFree(pNode), "delete failed");

			if(pTempNode)
				pNode = pTempNode;
			else
				pNode = NULL;
		}
	}

	if(pDrink->pDescriptors)
		ErrFatalDisplayIf(MemPtrFree(pDrink->pDescriptors), "delete failed");

	if(pDrink->sName)
		ErrFatalDisplayIf(MemPtrFree(pDrink->sName), "delete failed");

	if(pDrink->sDirections)
		ErrFatalDisplayIf(MemPtrFree(pDrink->sDirections), "delete failed");

	return error;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	DrinkPackedSize
// DESCRIPTION: determine the number of bytes a DrinkDB record uses
// PARAMETERS:
//
UInt32 DrinkPackedSize(DrinkPtr pDrink)
{
	UInt8				n;
	IngredientNodePtr	pNode;
	UInt32 				nSize;

	nSize = sizeof(UInt8);	// nIngredients
	pNode = pDrink->pIngredientList;
	for(n = 0; n < pDrink->nIngredients; n++)
	{
		nSize += sizeof(IngredientType);
		pNode = pNode->pNext;
	}

	nSize += sizeof(UInt8);	// nDescriptors;
	nSize += pDrink->nDescriptors;

	nSize += StrLen(pDrink->sName) + 1;
	nSize += StrLen(pDrink->sDirections) + 1;

	return nSize;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	FilterDrink
// DESCRIPTION: test a drink to see if it passes the filter requirements
// PARAMETERS:
//
Boolean FilterDrink(FilterType **filter, DrinkPtr pDrink)
{
	UInt8	nFtr, nDsc, nIng, nTest;
	char	*sTest;
	Boolean bInsert, bTestDrinkFilter;
	AppInfoBlockType info;

	// get the AppInfo Block, dont forget to free the memory!
	GetAppInfoBlock(&info);

	// whether or not the drink is added to the list.  assume it is and then test filters
	bInsert = true;

	// drink filters have an OR relationship, if any drink filter passes the drink is added
	bTestDrinkFilter = true;

	for(nFtr = 0; nFtr < nFILTERS; nFtr++)
	{
		// if this filter is turned off, go to the next one
		if(filter[nFtr]->action == OFF)
			continue;

		// test all available filters
		if(filter[nFtr]->target == DRINK && bTestDrinkFilter)
		{
			// test for the presence of the descriptor
			sTest = filter[nFtr]->psDescriptors[filter[nFtr]->nDescriptorSelection];
			nTest = StrPackedFind(sTest, info.pDrinkDescriptors, info.nDrinkDescriptors);
			for(nDsc = 0; nDsc < pDrink->nDescriptors; nDsc++)
			{
				if(filter[nFtr]->action == INCLUDE)
				{
					// stop checking on match
					if(pDrink->pDescriptors[nDsc] == nTest)
					{
						bInsert = true;
						bTestDrinkFilter = false;
						break;
					}

					// none of the descriptors match
					if(nDsc == pDrink->nDescriptors - 1)
					{
						bInsert = false;
						break;
					}
				}
				else if(filter[nFtr]->action == EXCLUDE)
				{
					// don't insert if we found a match
					if(pDrink->pDescriptors[nDsc] == nTest)
					{
						bInsert = false;
						break;
					}
				}
			}

			// if there are no descriptors, we know they weren't found
			if(pDrink->nDescriptors == 0)
				bInsert = false;

			// if we know the drink won't be added stop checking filters
			if(bInsert == false)
				continue;
		}
		else if(filter[nFtr]->target == INGREDIENT)
		{
			IngredientNodePtr pIngredient = pDrink->pIngredientList;
			sTest = filter[nFtr]->psIngredients[filter[nFtr]->nIngredientSelection];
			nTest = StrPackedFind(sTest, info.pIngredients, info.nIngredients);

			for(nIng = 0; nIng < pDrink->nIngredients; nIng++)
			{
				if(filter[nFtr]->action == INCLUDE)
				{
					// have we found a match for the ingredient name?
					if(pIngredient->ingredient.nName == nTest)
					{
						char *sD = filter[nFtr]->psDescriptors[filter[nFtr]->nDescriptorSelection];
						UInt8 nD = StrPackedFind(sD, info.pIngredientDescriptors, info.nIngredientDescriptors);

						// if descriptor set to "none" don't filter further,
						// otherwise check if there are descriptors to test
						if(!filter[nFtr]->bTypeChecked || filter[nFtr]->nDescriptorSelection + 1 == filter[nFtr]->nDescriptors)
							break;
						else if(pIngredient->ingredient.nDescriptors == 0)
						{
							bInsert = false;
							break;
						}

						if(pIngredient->ingredient.nDescriptor == nD)
							break;
					}
					else if(nIng == pDrink->nIngredients - 1)
					{
						bInsert = false;
						break;
					}
				}

				if(filter[nFtr]->action == EXCLUDE)
				{
					// have we found a match for the ingredient name?
					if(pIngredient->ingredient.nName == nTest)
					{
						char *sD = filter[nFtr]->psDescriptors[filter[nFtr]->nDescriptorSelection];
						UInt8 nD = StrPackedFind(sD, info.pIngredientDescriptors, info.nIngredientDescriptors);

						// if ingredient name matches and no descriptor specified, we exclude
						if(!filter[nFtr]->bTypeChecked) {
						    bInsert = false;
						    break;
      					}

						// descriptor  must match for exclude to happen
						if(filter[nFtr]->nDescriptorSelection == info.nIngredientDescriptors)
						{
							bInsert = false;
							break;
						}
						else if(pIngredient->ingredient.nDescriptor == nD)
						{
							bInsert = false;
							break;
						}
					}
				}
				pIngredient = pIngredient->pNext;
			}
		}

		// don't check the rest of the filters if this filter excludes the drink
		if(bInsert == false)
			break;
	}
	// clean up
	if(info.pDrinkDescriptors)
		ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
	if(info.pIngredientDescriptors)
		ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
	if(info.pIngredients)
		ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

	return bInsert;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetControlPtr
// DESCRIPTION:
// PARAMETERS:
//
ControlPtr GetControlPtr(FormPtr pForm, UInt16 nId)
{
   return FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, nId));
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetStringModal
// DESCRIPTION: popup a form to get a string from the user
// PARAMETERS:	sTitle = the text displayed in the form's title bar
//				sName = the string the user types
//
void GetStringModal(char* sTitle, char *sName)
{
	FormPtr pPrevForm = FrmGetActiveForm();
	FormPtr pForm = FrmInitForm(GetStringForm);
	FieldPtr pField;
	UInt16 nButton;
	char *s;

	FrmSetTitle(pForm, sTitle);
	FrmSetActiveForm(pForm);
	FrmSetFocus(pForm, FrmGetObjectIndex(pForm, GetStringField));

	// show the modal form, check that OK was tapped and get the string
	nButton = FrmDoDialog(pForm);
	if(nButton == DoneBtn)
	{
		pField = (FieldPtr)GetControlPtr(pForm, GetStringField);
		s = FldGetTextPtr(pField);
		if(s)
			StrCopy(sName, s);
	}
	else
		sName[0] = NULL;

	// switch back to previous form and clean up
	if(pPrevForm)
		FrmSetActiveForm(pPrevForm);
	FrmDeleteForm(pForm);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetAppInfoBlockPtr
// DESCRIPTION:
// PARAMETERS:
//
UInt8* GetAppInfoBlockPtr(void)
{
	UInt16 nCard;
	LocalID appInfoID;

	DmOpenDatabaseInfo(gDrinkDB, NULL, NULL, NULL, &nCard, NULL);
	appInfoID = DmGetAppInfoID(gDrinkDB);
	ErrFatalDisplayIf(appInfoID == 0, "error getting AppInfoBlock pointer");

	return (UInt8*)MemLocalIDToLockedPtr(appInfoID, nCard);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetAppInfoBlock
// DESCRIPTION:	stuffs an AppInfoBlockType struct with info from the database's
//				own AppInfoBlock
// PARAMETERS:
//
Err GetAppInfoBlock(AppInfoBlockPtr pInfo)
{
	char	*s;
	UInt8 	*pBlock, *p, *pIndex;
	UInt16	n, nSize;

	pBlock = GetAppInfoBlockPtr();
	pIndex = pBlock;

  	// get drink descriptors
	pInfo->nDrinkDescriptors = *(UInt8*)pIndex;
	pIndex++;
	p = pIndex;
	for(n = 0; n < pInfo->nDrinkDescriptors; n++)
		p += StrLen(p) + 1;
	nSize = p - pIndex;
	if(nSize > 0)
	{
		pInfo->pDrinkDescriptors = (char*)MemPtrNew(nSize * sizeof(char));
		MemMove(pInfo->pDrinkDescriptors, pIndex, nSize);
	}
	else
		pInfo->pDrinkDescriptors = NULL;

	// get ingredient descriptors
	pIndex = p;
	pInfo->nIngredientDescriptors = *(UInt8*)pIndex;
	pIndex++;
	p = pIndex;
	for(n = 0; n < pInfo->nIngredientDescriptors; n++)
		p += StrLen(p) + 1;
	nSize = p - pIndex;
	if(nSize > 0)
	{
		pInfo->pIngredientDescriptors = (char*)MemPtrNew(nSize * sizeof(char));
		MemMove(pInfo->pIngredientDescriptors, pIndex, nSize);
	}
	else
		pInfo->pIngredientDescriptors = NULL;

	// get ingredient names
	pIndex = p;
	pInfo->nIngredients = *(UInt8*)pIndex;
	pIndex++;
	p = pIndex;
	for(n = 0; n < pInfo->nIngredients; n++)
		p += StrLen(p) + 1;
	nSize = p - pIndex;
	if(nSize > 0)
	{
		pInfo->pIngredients = (char*)MemPtrNew(nSize * sizeof(char));
		MemMove(pInfo->pIngredients, pIndex, nSize);
	}
	else
		pInfo->pIngredients = NULL;

	MemPtrUnlock(pBlock);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetAppInfoBlockSize
// DESCRIPTION:	returns number of bytes needed to store the AppInfoBlockType
// PARAMETERS:
//
UInt32 GetAppInfoBlockSize(AppInfoBlockPtr pInfo)
{
	UInt32 nSize;

	nSize = 3;
	if(pInfo->pDrinkDescriptors)
		nSize += MemPtrSize(pInfo->pDrinkDescriptors);
	if(pInfo->pIngredientDescriptors)
		nSize += MemPtrSize(pInfo->pIngredientDescriptors);
	if(pInfo->pIngredients)
		nSize += MemPtrSize(pInfo->pIngredients);

	return nSize;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetRecordHandle
// DESCRIPTION: get a handle to a particular record in a database
// PARAMETERS:
//
Err GetRecordHandle(DmOpenRef *pDB, UInt16 nRecord, MemHandle *h)
{
	Err error = errNone;

	*h = DmQueryRecord(*pDB, nRecord);
	if(!h)
		error = DmGetLastErr();

	return error;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetFractionalLabel
// DESCRIPTION: return the ascii value of the fractional amount to be
//				displayed in the AddIngredient form based on the current
//				amount.
// PARAMETERS:
//
UInt8 GetFractionalLabel(UInt8 nAmount)
{
	UInt8	n = 0;

	switch(nAmount % 4)
	{
		case 0:
			n = '-';
			break;
		case 1:
			n = OneQuarterChar;
			break;
		case 2:
			n = OneHalfChar;
			break;
		case 3:
			n = ThreeQuartersChar;
			break;
	}

	return n;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetIngredients
// DESCRIPTION: get a string representation of the ingredient list
// PARAMETERS:
//
void GetIngredients(AppInfoBlockPtr pInfo, IngredientNodePtr pIngredientList,
	char ***psIngredients, UInt8 nIngredients)
{
	char sNew[MEDIUM_STRING];
	IngredientNodePtr pIndex;
	UInt16 nSize;
	UInt8 n;

	// bail if no ingredients
	if(nIngredients == 0)
	{
		psIngredients = NULL;
		return;
	}

	// allocate space for ingredients
	*psIngredients = (char**)MemPtrNew(nIngredients * sizeof(char*));

	pIndex = pIngredientList;
	n = 0;
	while(pIndex)
	{
		BuildIngredientString(pInfo, pIndex->ingredient, sNew);
		nSize = StrLen(sNew) + 1;
		(*psIngredients)[n] = (char*)MemPtrNew(nSize * sizeof(char));
		StrCopy((*psIngredients)[n], sNew);
		pIndex = pIndex->pNext;
		n++;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	GetDrinkList
// DESCRIPTION: fill the incoming array with drink names based on the current
// 				filter selections
// PARAMETERS:
//
void GetDrinkList(DmOpenRef *pDB, FilterType **filter, char ***psDrinks,
	UInt16 **pnRecordIndices, UInt16 *nDrinks)
{
	ListPtr		pList;
	DrinkPtr	pDrink;
	UInt16		n, nFtr, nIndex, nItems, nDrinksTotal, *pnTemp;
	UInt8 		*p;
	char		**psTemp;
	MemHandle	hRec;
	Boolean		bInsert;

	// create memory for a new drinks array
	nDrinksTotal = DmNumRecords(*pDB);
	if(nDrinksTotal)
	{
		pnTemp = (UInt16*)MemPtrNew(nDrinksTotal * sizeof(UInt16));
		psTemp = (char**)MemPtrNew(nDrinksTotal * sizeof(char*));
	}

	// search the database and get the filtered drink list
	nItems = 0;
	for(n = 0; n < nDrinksTotal; n++)
	{
		IngredientNodePtr pNode;

		// get the next drink record
		ErrFatalDisplayIf(GetRecordHandle(pDB, n, &hRec), "couldn't get record");
		p = (UInt8*)MemHandleLock(hRec);
		UnpackDrinkRecord(p, pDrink);
		MemHandleUnlock(hRec);

		// test it to see if it passes filter requirements
		/*
		if(filter[0]->target == OFF && filter[1]->target == OFF && filter[2]->target == OFF)
			bInsert = true;
		else
			bInsert = FilterDrink(filter, pDrink);
		*/

		// add the drink to the list if appropriate
		if(bInsert == true)
		{
			pnTemp[nItems] = n;
			psTemp[nItems] = (char*)MemPtrNew((StrLen(pDrink->sName) + 1) * sizeof(char));
			StrCopy(psTemp[nItems], pDrink->sName);
			nItems++;
		}

		ErrFatalDisplayIf(DrinkFree(pDrink), "error with free");
	}

	// copy the drink names from the temp storage to psDrinks
	*nDrinks = nItems;
	if(*nDrinks == 0)
	{
		*pnRecordIndices = NULL;
		*psDrinks = NULL;
	}
	else
	{
		*pnRecordIndices = (UInt16*)MemPtrNew((*nDrinks) * sizeof(UInt16));
		*psDrinks = (char**)MemPtrNew((*nDrinks) * sizeof(char*));
		for(n = 0; n < *nDrinks; n++)
		{
			(*pnRecordIndices)[n] = pnTemp[n];
			(*psDrinks)[n] = (char*)MemPtrNew((StrLen(psTemp[n]) + 1) * sizeof(char));
			StrCopy((*psDrinks)[n], psTemp[n]);
		}
	}

	StrArrayFree(psTemp, nDrinksTotal);
	if(psTemp)
	{
		ErrFatalDisplayIf(MemPtrFree(psTemp), "delete failed");
		psTemp = NULL;
	}
	if(pnTemp)
	{
		ErrFatalDisplayIf(MemPtrFree(pnTemp), "error with free");
		pnTemp = NULL;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	RandInt
// DESCRIPTION:
// PARAMETERS:
//
Int16 RandInt(const int nMax)
{
	return(SysRandom(0) % nMax);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	ScrollLines
// DESCRIPTION:
// PARAMETERS:
//
void ScrollLines(FieldPtr pField, ScrollBarPtr pScroll, const Int16 nLines,
	const Boolean redraw)
{
	if(nLines < 0)
		FldScrollField(pField, -nLines, winUp);
	else
		FldScrollField(pField, nLines, winDown);

	if(FldGetNumberOfBlankLines(pField) && (nLines < 0) || redraw)
		UpdateScrollbar(pField, pScroll);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	SetFieldText
// DESCRIPTION:
// PARAMETERS:
//
void SetFieldText(FormPtr pForm, UInt16 nId, char* s)
{
   FieldPtr pField = (FieldPtr)GetControlPtr(pForm, nId);
   if(pField == NULL)
      return;
   FldSetTextPtr(pField, s);
   FldRecalculateField(pField, true);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	SetScreenMode
// DESCRIPTION: sets the screen mode to the highest supported depth
// PARAMETERS:
//
UInt32 SetScreenMode(void)
{
	UInt32 nDepth;

	// Initialize the screen with the highest available color depth
	WinScreenMode(winScreenModeGetSupportedDepths, NULL, NULL, &nDepth, NULL);
	if(nDepth & 128)
		nDepth = 8;		// 8bit color (256 colors)
	else if(nDepth & 8)
		nDepth = 4;		// 4bit gray (16 shades of gray)
	else if(nDepth & 2)
		nDepth = 2;		// 2bit gray (4 shades of gray)

	WinScreenMode(winScreenModeSet, NULL, NULL, &nDepth, NULL);

	return nDepth;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	StrPackedCopy
// DESCRIPTION: copy a series of null-terminated strings into a string array
// PARAMETERS:
//
void StrPackedCopy(char *pSource, char **psDest, UInt16 nItems)
{
	UInt16 n, nSize;
	char *p;

	p = pSource;
	for(n = 0; n < nItems; n++)
	{
		nSize = StrLen(p) + 1;
		psDest[n] = (char*)MemPtrNew(nSize * sizeof(char));
		StrCopy(psDest[n], p);
		p += nSize;
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	StrArrayFree
// DESCRIPTION: deallocate memory associated with a string array
// PARAMETERS:
//
void StrArrayFree(char *psArray[], UInt16 nItems)
{
	UInt16 n;

	// bail if nothing to do
	if(psArray == NULL)
		return;

	// free up memory allocated for the array
	for(n = 0; n < nItems; n++)
	{
		if(psArray[n] != NULL)
		{
			ErrFatalDisplayIf(MemPtrFree(psArray[n]), "delete failed");
			psArray[n] = NULL;
		}

	}
//	ErrFatalDisplayIf(MemPtrFree(psArray), "delete failed");
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION:	StrArrayInsert
// DESCRIPTION: return an array built from a packed set of strings with the
//				string sNew	inserted at index nPos
// PARAMETERS:	sNew = the string to insert
//				nPos = the index where sNew should end up
//				pSource =
//				nItems = the number of items pre-insertion
//
char** StrArrayInsert(char *sNew, UInt8 nPos, char *pSource, UInt8 *nItems)
{
	UInt8 n, nLength;
	char **psArray;
	char *sTest;

	*nItems = *nItems + 1;
	psArray = (char**)MemPtrNew(*nItems * sizeof(char*));

	sTest = pSource;
	for(n = 0; n < *nItems; n++)
	{
		if(n == nPos)
		{
			nLength = StrLen(sNew) + 1;
			psArray[n] = (char*)MemPtrNew(nLength * sizeof(char));
			StrCopy(psArray[n], sNew);
//			sTest += nLength;
		}
		else
		{
			nLength = StrLen(sTest) + 1;
			psArray[n] = (char*)MemPtrNew(nLength * sizeof(char));
			StrCopy(psArray[n], sTest);
			sTest += nLength;
		}
	}

	return psArray;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	StrArraySwap
// DESCRIPTION: swap two strings in the incoming array
// PARAMETERS:
//
void StrArraySwap(char ***ppsArray, Int16 i, Int16 j)
{
	char *sTemp;

	sTemp = (*ppsArray)[i];
	(*ppsArray)[i] = (*ppsArray)[j];
	(*ppsArray)[j] = sTemp;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	StrArraySort
// DESCRIPTION: reorder the incoming array of strings alphabetically
// PARAMETERS:
//
void StrArraySort(char ***ppsArray, Int16 nLeft, Int16 nRight)
{
	UInt8 n, nLast;

	// do nothing if sorting less than 2 elements
	if(nLeft >= nRight)
		return;

	StrArraySwap(ppsArray, nLeft, (nLeft + nRight) / 2);
	nLast = nLeft;
	for(n = nLeft + 1; n <= nRight; n++)
	{
		if(StrCompare((*ppsArray)[n], (*ppsArray)[nLeft]) < 0)
			StrArraySwap(ppsArray, ++nLast, n);
	}

	StrArraySwap(ppsArray, nLeft, nLast);
	StrArraySort(ppsArray, nLeft, nLast - 1);
	StrArraySort(ppsArray, nLast + 1, nRight);
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	StrPackedFind
// DESCRIPTION: test packed set of strings for a match of the input string,
// 				and return the index of the match or -1 for not present
// PARAMETERS:
//
Int16 StrPackedFind(char *sInput, char *pSource, UInt16 nItems)
{
	UInt8 n, nLength;
	Int16 nResult;
	char *sTest;

	nResult = -1;
	sTest = pSource;
	for(n = 0; n < nItems; n++)
	{
		nLength = StrLen(sTest) + 1;
		if(StrCompare(sTest, sInput) == 0)
		{
			nResult = n;
			break;
		}
		sTest += nLength;
	}

	return nResult;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	UnpackDrinkRecord
// DESCRIPTION: create a drink type from a packed drink record
// PARAMETERS:
//
Err UnpackDrinkRecord(UInt8 *p, DrinkPtr pDrink)
{
	UInt16	n, nSize;
	IngredientNode *pNode, *pBefore;
	Err 	error = errNone;

	pDrink->nIngredients = *(UInt8*)p;
	p++;

	for(n = 0; n < pDrink->nIngredients; n++)
	{
		pNode = (IngredientNodePtr)MemPtrNew(sizeof(IngredientNode));
		MemMove(&pNode->ingredient, p, sizeof(IngredientType));
		p += sizeof(IngredientType);

		if(n == 0)
		{
			pNode->pNext = NULL;
			pNode->pPrevious = NULL;
			pDrink->pIngredientList = pNode;
			pBefore = pNode;
		}
		else
		{
			pNode->pNext = NULL;
			pNode->pPrevious = pBefore;
			pBefore->pNext = pNode;
			pBefore = pNode;
		}
	}

	pDrink->nDescriptors = *(UInt8*)p;
	p++;

	nSize = pDrink->nDescriptors * sizeof(UInt8);
	if(nSize)
	{
		pDrink->pDescriptors = (UInt8*)MemPtrNew(nSize);
		MemMove(pDrink->pDescriptors, p, nSize);
		p += nSize;
	}
	else
		pDrink->pDescriptors = NULL;

	nSize = (StrLen(p) + 1) * sizeof(char);
	pDrink->sName = (char*)MemPtrNew(nSize);
	StrCopy(pDrink->sName, p);
	p += nSize;

	nSize = (StrLen(p) + 1) * sizeof(char);
	pDrink->sDirections = (char*)MemPtrNew(nSize);
	StrCopy(pDrink->sDirections, p);
	p += nSize;

	return error;
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	UpdateArrowState
// DESCRIPTION: update the UI and enabled/disabled state of the up/down arrows
// PARAMETERS:
//
void UpdateArrowState(FormPtr pForm, IngredientNodePtr pIngredientList,
	char *sUpLabel, char *sDownLabel)
{
	ListPtr pList = (ListPtr)GetControlPtr(pForm, IngredientList);
	void *pControl;
	UInt8 n;
	Int16 nSelect = LstGetSelection(pList);
	IngredientNodePtr pNode;

	if(nSelect == noListSelection)
	{
		pControl = (ControlPtr)GetControlPtr(pForm, UpBtn);
		CtlSetEnabled(pControl, false);
		sUpLabel[0] = UpArrowDisabled;
		CtlSetLabel(pControl, sUpLabel);

		pControl = (ControlPtr)GetControlPtr(pForm, DownBtn);
		CtlSetEnabled(pControl, false);
		sDownLabel[0] = DownArrowDisabled;
		CtlSetLabel(pControl, sDownLabel);
	}
	else
	{
		pNode = pIngredientList;
		for(n = 0; n < nSelect; n++)
			pNode = pNode->pNext;

		pControl = (ControlPtr)GetControlPtr(pForm, UpBtn);
		if(pNode->pPrevious)
		{
			CtlSetEnabled(pControl, true);
			sUpLabel[0] = UpArrowEnabled;
		}
		else
		{
			CtlSetEnabled(pControl, false);
			sUpLabel[0] = UpArrowDisabled;
		}
		CtlSetLabel(pControl, sUpLabel);

		pControl = (ControlPtr)GetControlPtr(pForm, DownBtn);
		if(pNode->pNext)
		{
			CtlSetEnabled(pControl, true);
			sDownLabel[0] = DownArrowEnabled;
		}
		else
		{
			CtlSetEnabled(pControl, false);
			sDownLabel[0] = DownArrowDisabled;
		}
		CtlSetLabel(pControl, sDownLabel);
	}
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	UpdateFilterUI
// DESCRIPTION: set up the filter form to reflect the filter settings
// PARAMETERS:
//
void UpdateFilterUI(FormPtr pForm, FilterType **filter, UInt8 nFilter)
{
	void		*pObj;
	ListPtr		pList;
	UInt8		n, nIndex;
	char 		*s;
	AppInfoBlockType info;

	// get the AppInfo Block, dont forget to free the memory!
	GetAppInfoBlock(&info);

	// check the current filter's target and adjust the UI accordingly
	n = nFilter - 1;
	if(filter[n]->action == OFF)
	{
		// unhide the appropriate form elements
		pObj = GetControlPtr(pForm, FilterActionOffPB);
		CtlSetValue(pObj, 1);
		pObj = GetControlPtr(pForm, FilterActionInclude);
		CtlSetValue(pObj, 0);
		pObj = GetControlPtr(pForm, FilterActionExclude);
		CtlSetValue(pObj, 0);
		pObj = GetControlPtr(pForm, FilterDrinkPB);
		CtlSetUsable(pObj, false);
		pObj = GetControlPtr(pForm, FilterIngredientPB);
		CtlSetUsable(pObj, false);
		pObj = GetControlPtr(pForm, FilterDescriptorPopup);
		CtlSetUsable(pObj, false);
		pObj = GetControlPtr(pForm, FilterDescriptorList);
		CtlSetUsable(pObj, false);
		pObj = GetControlPtr(pForm, FilterTypeCheckbox);
		CtlSetUsable(pObj, false);
		FrmHideObject(pForm, FrmGetObjectIndex(pForm, FilterIngredientList));
		FrmCopyLabel(pForm, FilterTypeLabel, "    ");
	}
	else
	{
		pObj = GetControlPtr(pForm, FilterActionOffPB);
		CtlSetValue(pObj, 0);

		if(filter[n]->action == INCLUDE)
		{
			pObj = GetControlPtr(pForm, FilterActionInclude);
			CtlSetValue(pObj, 1);
			pObj = GetControlPtr(pForm, FilterActionExclude);
			CtlSetValue(pObj, 0);
		}

		if(filter[n]->action == EXCLUDE)
		{
			pObj = GetControlPtr(pForm, FilterActionExclude);
			CtlSetValue(pObj, 1);
			pObj = GetControlPtr(pForm, FilterActionInclude);
			CtlSetValue(pObj, 0);
		}

		switch(filter[n]->target)
		{
		case DRINK:
			// erase the old descriptors array
			StrArrayFree(filter[n]->psDescriptors, filter[n]->nDescriptors);
			if(filter[n]->psDescriptors)
			{
				ErrFatalDisplayIf(MemPtrFree(filter[n]->psDescriptors), "delete failed");
				filter[n]->psDescriptors = NULL;
			}

			// fill the new descriptors array
			filter[n]->nDescriptors = info.nDrinkDescriptors;
			filter[n]->psDescriptors = (char**)MemPtrNew(filter[n]->nDescriptors *
				sizeof(char*));

			s = info.pDrinkDescriptors;
			for(nIndex = 0; nIndex < info.nDrinkDescriptors; nIndex++)
			{
				filter[n]->psDescriptors[nIndex] = MemPtrNew((StrLen(s) + 1) *
					sizeof(char));
				StrCopy(filter[n]->psDescriptors[nIndex], s);
				s += StrLen(s) + 1;
			}
			StrArraySort(&(filter[n]->psDescriptors), 0, filter[n]->nDescriptors - 1);

			// activate and deactivate the appropriate ui elements
			//
			pObj = GetControlPtr(pForm, FilterDrinkPB);
			CtlSetValue(pObj, 1);
			pObj = GetControlPtr(pForm, FilterIngredientPB);
			CtlSetValue(pObj, 0);

			pObj = GetControlPtr(pForm, FilterDrinkPB);
			CtlSetUsable(pObj, true);
			pObj = GetControlPtr(pForm, FilterIngredientPB);
			CtlSetUsable(pObj, true);

			pObj = GetControlPtr(pForm, FilterTypeCheckbox);
			CtlSetUsable(pObj, false);

			FrmCopyLabel(pForm, FilterTypeLabel, "    ");

			pObj = GetControlPtr(pForm, FilterDescriptorPopup);
			CtlSetUsable(pObj, false);

			pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
			LstSetListChoices(pList, filter[n]->psDescriptors, filter[n]->nDescriptors);
			LstSetSelection(pList, filter[n]->nDescriptorSelection);
			LstSetTopItem(pList, filter[n]->nDescriptorSelection);
			FrmShowObject(pForm, FrmGetObjectIndex(pForm, FilterIngredientList));

			break;

		case INGREDIENT:
			// erase the old descriptors array
			StrArrayFree(filter[n]->psDescriptors, filter[n]->nDescriptors);
			if(filter[n]->psDescriptors)
			{
				ErrFatalDisplayIf(MemPtrFree(filter[n]->psDescriptors), "delete failed");
				filter[n]->psDescriptors = NULL;
			}

			// fill in the new descriptors array
			filter[n]->nDescriptors = info.nIngredientDescriptors + 1; //plus one for "none" selection
			filter[n]->psDescriptors = (char**)MemPtrNew(filter[n]->nDescriptors * sizeof(char*));

			s = (char*)info.pIngredientDescriptors;
			for(nIndex = 0; nIndex < info.nIngredientDescriptors; nIndex++)
			{
				filter[n]->psDescriptors[nIndex] = MemPtrNew((StrLen(s) + 1) *
					sizeof(char));
				StrCopy(filter[n]->psDescriptors[nIndex], s);
				s += StrLen(s) + 1;
			}
			StrArraySort(&(filter[n]->psDescriptors), 0, filter[n]->nDescriptors - 2);
			filter[n]->psDescriptors[nIndex] = MemPtrNew((StrLen(gsNone) + 1) * sizeof(char));
			StrCopy(filter[n]->psDescriptors[nIndex], gsNone);

			// erase the old ingredients array
			StrArrayFree(filter[n]->psIngredients, filter[n]->nIngredients);
			if(filter[n]->psIngredients)
			{
				ErrFatalDisplayIf(MemPtrFree(filter[n]->psIngredients), "delete failed");
				filter[n]->psIngredients = NULL;
			}

			// fill in the new ingredients array
			filter[n]->nIngredients = info.nIngredients;
			filter[n]->psIngredients = (char**)MemPtrNew(filter[n]->nIngredients *
				sizeof(char*));

			s = (char *)info.pIngredients;
			for(nIndex = 0; nIndex < info.nIngredients; nIndex++)
			{
				filter[n]->psIngredients[nIndex] = MemPtrNew((StrLen(s) + 1) *
					sizeof(char));
				StrCopy(filter[n]->psIngredients[nIndex], s);
				s += StrLen(s) + 1;
			}
			StrArraySort(&(filter[n]->psIngredients), 0, filter[n]->nIngredients - 1);

			// activate and deactivate the appropriate ui elements
			//
			pObj = GetControlPtr(pForm, FilterActionOffPB);
			CtlSetValue(pObj, 0);

			pObj = GetControlPtr(pForm, FilterIngredientPB);
			CtlSetValue(pObj, 1);
			pObj = GetControlPtr(pForm, FilterDrinkPB);
			CtlSetValue(pObj, 0);

			pObj = GetControlPtr(pForm, FilterDrinkPB);
			CtlSetUsable(pObj, true);
			pObj = GetControlPtr(pForm, FilterIngredientPB);
			CtlSetUsable(pObj, true);

			FrmCopyLabel(pForm, FilterTypeLabel, "Type:");
			pObj = GetControlPtr(pForm, FilterTypeCheckbox);
			CtlSetUsable(pObj, true);

			if(filter[n]->bTypeChecked)
			{
				CtlSetValue(pObj, 1);
				pObj = GetControlPtr(pForm, FilterDescriptorPopup);
				CtlSetUsable(pObj, true);
				CtlSetLabel(pObj, filter[n]->psDescriptors[filter[n]->nDescriptorSelection]);
   			}
			else
			{
				CtlSetValue(pObj, 0);
				pObj = GetControlPtr(pForm, FilterDescriptorPopup);
				CtlSetUsable(pObj, false);
   			}

			pList = (ListPtr)GetControlPtr(pForm, FilterDescriptorList);
			LstSetListChoices(pList, filter[n]->psDescriptors, filter[n]->nDescriptors);
			LstSetSelection(pList, noListSelection);

			pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
			LstSetListChoices(pList, filter[n]->psIngredients, filter[n]->nIngredients);
			LstSetSelection(pList, filter[n]->nIngredientSelection);
			LstSetTopItem(pList, filter[n]->nIngredientSelection);
			FrmShowObject(pForm, FrmGetObjectIndex(pForm, FilterIngredientList));
			break;
		}

	}

	if(info.pDrinkDescriptors)
		ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
	if(info.pIngredientDescriptors)
		ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
	if(info.pIngredients)
		ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");
}

/////////////////////////////////////////////////////////////////////////////
// FUNCTION: 	UpdateScrollbar
// DESCRIPTION:
// PARAMETERS:
//
void UpdateScrollbar(FieldPtr pField, ScrollBarPtr pScroll)
{
	UInt16 curPos, textHeight, fieldHeight, maxValue;

	FldGetScrollValues(pField, &curPos, &textHeight, &fieldHeight);
	if(textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if(curPos)
		maxValue = curPos;
	else
		maxValue = 0;

	SclSetScrollBar(pScroll, curPos, 0, maxValue, fieldHeight - 1);
}

// EOF
