/*  Minibar - minibar.c "main source file"
    Copyright (C) 2002-2005 Ray Vanderborght (depsypher@users.sourceforge.net)

    This file is part of Minibar.

    Minibar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Minibar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Minibar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Notes for Sourceforge
 *
 * CVS:
 *   :extssh:depsypher@cvs.sourceforge.net:/cvsroot/minibar
 *
 * File Release:
 *   1. FTP to upload.sourceforge.net
 *   2. Login as "anonymous"
 *   3. Use your e-mail address as the password for this login
 *   4. Set your client to binary mode ("bin" on command-line clients)
 *   5. Change your current directory to /incoming ("cd /incoming")
 *   6. Upload the desired files for the release ("put filename")
 *   7. From the project page go to Admin>File Releases>Add Release
 *
 * Changes in 2.2
 *   1. Changed AppFileCreatorId to match the database id
 *   2. Fixed Treo Filter crash bug thanks to Maroy Melin
 *
 * Changes in 2.1
 *   1. Fixed bug preventing "Browse Drinks" form from appearing in PalmOS5
 *   2. Fixed Fatal Exception bug in "Add Drink" on PalmOS4.1
 *   3. Added support for up & down keys in Recipe form (scrolling recipe text)
 *   4. Properly closes the database when exiting the app
 *   5. Filter UI improvements
 *
 * Features Needed
 *   1. Edit Drink ability
 *   2. Intelligent Ingredient Descriptors in Filter Form
 *   3. Externalize strings for internationalization
 */

#include <PalmOS.h>
#include <PalmOSGlue.h>
#include "minibar.h"
#include "protos.h"

//===========================================================================
//	Globals...
//===========================================================================

extern UInt16			gPreviousForm;
extern DmOpenRef		gDrinkDB;
extern IngredientType	gIngredient;
extern DrinkType		gDrink;
extern FilterPtr		*gFilterSettings;
extern StringArrayType	gDrinkDescriptors;
extern StringArrayType	gIngredientDescriptors;
extern StringArrayType	gIngredients;
extern char				*gsNone;
extern char				*gsFilterTitle;
extern UInt16			gnBrowseDrinks;
extern char				**gpsBrowseDrinks;
extern UInt16			*gpnBrowseDrinkIndices;
extern UInt8			gnCurrentFilter;
extern Boolean			gUpdateDrinkList;

//===========================================================================
//	Event Handlers...
//===========================================================================

static Boolean MainFormEventH(const EventPtr event)
{
	FormPtr 		pForm = FrmGetActiveForm();
	void			*pObj;
	Boolean			isHandled = false;
	UInt32			n, nIndex;
	char			*s, *sRecord;
	MemHandle		hRec;
	ListPtr			pList;

	switch(event->eType)
	{
		case frmOpenEvent:
			gPreviousForm = MainForm;
			FrmDrawForm(pForm);
			isHandled = true;
			break;

		case ctlSelectEvent:
			if(event->data.ctlEnter.controlID == DrinksBtn)
			{
				FrmGotoForm(DrinksForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == RandomBtn)
			{
				UInt8 *p;

				n = RandInt(DmNumRecords(gDrinkDB));
				ErrFatalDisplayIf(GetRecordHandle(&gDrinkDB, n, &hRec),
					"couldn't get record");
				p = (UInt8*)MemHandleLock(hRec);
				UnpackDrinkRecord(p, &gDrink);
				MemHandleUnlock(hRec);

				FrmGotoForm(RecipeForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == InfoBtn)
				FrmHelp(MainHelpText);
			else if(event->data.ctlEnter.controlID == AboutBtn)
				FrmHelp(MainAboutText);
			break;

		case menuEvent:
			if(event->data.menu.itemID == MainBrowse)
			{
				FrmGotoForm(DrinksForm);
			}
			else if(event->data.menu.itemID == MainSurprise)
			{
				UInt8 *p;
				n = RandInt(DmNumRecords(gDrinkDB));
				ErrFatalDisplayIf(GetRecordHandle(&gDrinkDB, n, &hRec),
					"couldn't get record");
				p = (UInt8*)MemHandleLock(hRec);
				UnpackDrinkRecord(p, &gDrink);
				MemHandleUnlock(hRec);

				FrmGotoForm(RecipeForm);
			}
			else if(event->data.menu.itemID == MainHelp)
			{
				FrmHelp(MainHelpText);
			}
			else if(event->data.menu.itemID == MainAbout)
			{
				FrmHelp(MainAboutText);
			}

			break;
	}

	return isHandled;
}

static Boolean AddDrinkFormEventH(const EventPtr event)
{
	FormPtr 				pForm = FrmGetActiveForm();
	ScrollBarPtr			pScroll;
	FieldPtr				pField;
	ListPtr					pList;
	ControlPtr				pControl;
	static char				sUpLabel[2], sDownLabel[2];
	static char				**psIngredients;
	static UInt8			nIngredients;
	static IngredientNode	*pIngredientList;
	IngredientNode			*pNew, *pNode;
	char					sNew[MEDIUM_STRING];
	char					*pIndex, *pTemp;
	UInt8					*p;
	Int16 					nSelect;
	UInt16					n, nSize;
	Boolean					isHandled;
	AppInfoBlockType		info;

	isHandled = false;
	switch(event->eType)
	{
		case winEnterEvent:
			// we're either entering this form for the first time or returning
			// from the AddIngredient form.  Add new ingredient if appropriate
			if(gIngredient.nAmount == 0)
				break;

			// add the new ingredient to the end of the list
			pNew = (IngredientNodePtr)MemPtrNew(sizeof(IngredientNode));
			MemMove(&(pNew->ingredient), &gIngredient, sizeof(gIngredient));

			if(pIngredientList == NULL)
			{
				pIngredientList = pNew;
				pIngredientList->pNext = NULL;
				pIngredientList->pPrevious = NULL;
			}
			else
			{
				pNode = pIngredientList;
				while(pNode->pNext)
					pNode = pNode->pNext;

				pNew->pNext = NULL;
				pNew->pPrevious = pNode;
				pNode->pNext = pNew;
			}

			GetAppInfoBlock(&info);
			StrArrayFree(psIngredients, nIngredients);
			if(psIngredients)
			{
				ErrFatalDisplayIf(MemPtrFree(psIngredients), "delete failed");
				psIngredients = NULL;
			}
			nIngredients++;
			GetIngredients(&info, pIngredientList, &psIngredients, nIngredients);

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			gIngredient.nAmount = 0;
			pList = (ListPtr)GetControlPtr(pForm, IngredientList);
			LstSetListChoices(pList, psIngredients, nIngredients);
			UpdateArrowState(pForm, pIngredientList, sUpLabel, sDownLabel);
			LstDrawList(pList);			// redraw
			break;

		case frmOpenEvent:
			pScroll = (ScrollBarPtr)GetControlPtr(pForm, DirectionsScroll);
			pField = (FieldPtr)GetControlPtr(pForm, DirectionsField);
			pList = (ListPtr)GetControlPtr(pForm, IngredientList);

			sUpLabel[1] = NULL;
			sDownLabel[1] = NULL;
			nIngredients = 0;
			pIngredientList = NULL;

			gDrink.nIngredients = 0;
			gDrink.nDescriptors = 0;

			pControl = (ControlPtr)GetControlPtr(pForm, UpBtn);
			CtlSetEnabled(pControl, false);
			pControl = (ControlPtr)GetControlPtr(pForm, DownBtn);
			CtlSetEnabled(pControl, false);

			FrmSetFocus(pForm, FrmGetObjectIndex(pForm, DrinkNameField));
			LstSetListChoices(pList, psIngredients, nIngredients);
			LstSetSelection(pList, noListSelection);
			UpdateScrollbar(pField, pScroll);
			FrmDrawForm(pForm);

			isHandled = true;
			break;

		case frmCloseEvent:
			StrArrayFree(psIngredients, nIngredients);
			if(psIngredients)
			{
				ErrFatalDisplayIf(MemPtrFree(psIngredients), "delete failed");
				psIngredients = NULL;
			}

			pNode = pIngredientList;
			while(pNode)
			{
				pNode = pIngredientList->pNext;
				ErrFatalDisplayIf(MemPtrFree(pIngredientList),
					"error deleting");
				pIngredientList = pNode;
			}
			pIngredientList = NULL;
			break;

		case menuEvent:
			if(event->data.menu.itemID == AddIngredient)
				FrmPopupForm(AddIngredientForm);
			break;

		case lstSelectEvent:
			UpdateArrowState(pForm, pIngredientList, sUpLabel, sDownLabel);
			isHandled = true;
			break;

		case fldChangedEvent:
			pScroll = (ScrollBarPtr)GetControlPtr(pForm, DirectionsScroll);
			pField = (FieldPtr)GetControlPtr(pForm, DirectionsField);
			UpdateScrollbar(pField, pScroll);
			isHandled = true;
			break;

		case sclRepeatEvent:
			pScroll = (ScrollBarPtr)GetControlPtr(pForm, DirectionsScroll);
			pField = (FieldPtr)GetControlPtr(pForm, DirectionsField);
			ScrollLines(pField, pScroll, event->data.sclRepeat.newValue -
				event->data.sclRepeat.value, false);
			break;

		case ctlRepeatEvent:
			// switch an ingredient with the one before it in the list
			if(event->data.ctlRepeat.controlID == UpBtn)
			{
				IngredientNode *pBefore, *pAfter;
				pList = (ListPtr)GetControlPtr(pForm, IngredientList);
				nSelect = LstGetSelection(pList);

				pNode = pIngredientList;
				for(n = 0; n < nSelect; n++)
					pNode = pNode->pNext;

				pBefore = pNode->pPrevious;

				if(pNode->pNext)
				{
					pAfter = pNode->pNext;
					pBefore->pNext = pAfter;
				}
				else
				{
					pAfter = NULL;
					pBefore->pNext = NULL;
				}
				pNode->pNext = pBefore;

				if(pBefore->pPrevious)
				{
					pBefore->pPrevious->pNext = pNode;
					pNode->pPrevious = pBefore->pPrevious;
				}
				else
				{
					pNode->pPrevious = NULL;
					pIngredientList = pNode;
				}
				pBefore->pPrevious = pNode;

				if(pAfter)
					pAfter->pPrevious = pBefore;

				nSelect--;	// make sure the selected item moves as well
			}

			// switch the ingredient with the one after it in the list
			if(event->data.ctlRepeat.controlID == DownBtn)
			{
				IngredientNode *pBefore, *pAfter;
				pList = (ListPtr)GetControlPtr(pForm, IngredientList);
				nSelect = LstGetSelection(pList);

				pNode = pIngredientList;
				for(n = 0; n < nSelect; n++)
					pNode = pNode->pNext;

				pAfter = pNode->pNext;
				if(pNode->pPrevious)
				{
					pBefore = pNode->pPrevious;
					pAfter->pPrevious = pBefore;
				}
				else
				{
					pBefore = NULL;
					pAfter->pPrevious = NULL;
					pIngredientList = pAfter;
				}

				pNode->pPrevious = pAfter;
				if(pAfter->pNext)
				{
					pAfter->pNext->pPrevious = pNode;
					pNode->pNext = pAfter->pNext;
				}
				else
					pNode->pNext = NULL;

				pAfter->pNext = pNode;

				if(pBefore)
					pBefore->pNext = pAfter;

				nSelect++;	// make sure the selected item moves as well
			}

			GetAppInfoBlock(&info);
			StrArrayFree(psIngredients, nIngredients);
			if(psIngredients)
			{
				ErrFatalDisplayIf(MemPtrFree(psIngredients), "delete failed");
				psIngredients = NULL;
			}
			GetIngredients(&info, pIngredientList, &psIngredients, nIngredients);

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			pList = (ListPtr)GetControlPtr(pForm, IngredientList);
			LstSetListChoices(pList, psIngredients, nIngredients);
			LstSetSelection(pList, nSelect);
			UpdateArrowState(pForm, pIngredientList, sUpLabel, sDownLabel);

			LstDrawList(pList);	// redraw
			break;

		case keyDownEvent:
			if (event->data.keyDown.chr == vchrPageUp)
			{
				pList = (ListPtr)GetControlPtr(pForm, IngredientList);
				LstScrollList(pList, winUp, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else if (event->data.keyDown.chr == vchrPageDown)
			{
			 	pList = (ListPtr)GetControlPtr(pForm, IngredientList);
				LstScrollList(pList, winDown, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			break;

		case ctlSelectEvent:
			if(event->data.ctlEnter.controlID == AddBtn)
			{
				FrmPopupForm(AddIngredientForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == RemoveBtn)
			{
				pList = (ListPtr)GetControlPtr(pForm, IngredientList);
				nSelect = LstGetSelection(pList);
				if(nSelect == noListSelection)
				{
					DisplayString("Please select an ingredient to remove.");
					isHandled = true;
					break;
				}

				pNode = pIngredientList;
				for(n = 0; n < nSelect; n++)
					pNode = pNode->pNext;

				if(pNode->pPrevious)
					pNode->pPrevious->pNext = pNode->pNext;
				else
					pIngredientList = pNode->pNext;

				if(pNode->pNext)
					pNode->pNext->pPrevious = pNode->pPrevious;

				ErrFatalDisplayIf(MemPtrFree(pNode), "error with free");

				nIngredients--;
				if(nIngredients == 0)
				{
					pIngredientList = NULL;
					nSelect = noListSelection;
				}

				// move the selection if it was the last one in the list
				if(nSelect > nIngredients - 1)
					nSelect = nIngredients - 1;

				GetAppInfoBlock(&info);
				GetIngredients(&info, pIngredientList, &psIngredients,
					nIngredients);
				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				LstSetListChoices(pList, psIngredients, nIngredients);
				LstSetSelection(pList, nSelect);
				UpdateArrowState(pForm, pIngredientList, sUpLabel, sDownLabel);
				LstDrawList(pList);			// redraw

				gDrink.nIngredients--;
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == DrinkDescriptBtn)
			{
				FrmPopupForm(AddDescriptorForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == DoneBtn)
			{
				pField = (FieldPtr)GetControlPtr(pForm, DrinkNameField);
				gDrink.sName = FldGetTextPtr(pField);
				if(gDrink.sName == NULL)
				{
					DisplayString("Please enter a Drink Name.");
					isHandled = true;
					break;
				}

				pField = (FieldPtr)GetControlPtr(pForm, DirectionsField);
				gDrink.sDirections = FldGetTextPtr(pField);
				if(gDrink.sDirections == NULL)
				{
					DisplayString("Please enter Drink Directions.");
					isHandled = true;
					break;
				}

				if(pIngredientList == NULL)
				{
					DisplayString("Please enter the Drink Ingredients.");
					isHandled = true;
					break;
				}

				gDrink.pIngredientList = pIngredientList;

				// build the list of drink descriptors from the string array
				GetAppInfoBlock(&info);
				gDrink.nDescriptors = gDrinkDescriptors.nItems;
				if(gDrink.nDescriptors == 0)
				{
					gDrink.pDescriptors = (UInt8*)MemPtrNew(1 * sizeof(UInt8));
					gDrink.pDescriptors[0] = 0;
				}
				else
				{
					gDrink.pDescriptors = (UInt8*)MemPtrNew(gDrink.nDescriptors * sizeof(UInt8));
					for(n = 0; n < gDrink.nDescriptors; n++)
					{
						gDrink.pDescriptors[n] = (UInt8)StrPackedFind(gDrinkDescriptors.psArray[n],
							info.pDrinkDescriptors, info.nDrinkDescriptors);
					}
				}

				// add the new record to the database
				ErrFatalDisplayIf(AddDrinkRecord(dmMaxRecordIndex, &gDrink),
					"error adding record");

				// sort the database alphabetically by drink name
				if(DmNumRecords(gDrinkDB) > 1)
					DmInsertionSort(gDrinkDB, (DmComparF*)CompareDrinkRecords, 0);

				// clean up
				if(gDrink.pDescriptors)
					ErrFatalDisplayIf(MemPtrFree(gDrink.pDescriptors), "delete failed");
				StrArrayFree(gDrinkDescriptors.psArray, gDrinkDescriptors.nItems);
				if(gDrinkDescriptors.psArray)
				{
					ErrFatalDisplayIf(MemPtrFree(gDrinkDescriptors.psArray), "delete failed");
					gDrinkDescriptors.psArray = NULL;
				}
				gDrinkDescriptors.nItems = 0;

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				// erase the current drinks list so it will be rebuilt
				StrArrayFree(gpsBrowseDrinks, gnBrowseDrinks);
				if(gpsBrowseDrinks)
				{
					ErrFatalDisplayIf(MemPtrFree(gpsBrowseDrinks), "delete failed");
					gpsBrowseDrinks = NULL;
				}
				if(gpnBrowseDrinkIndices)
				{
					ErrFatalDisplayIf(MemPtrFree(gpnBrowseDrinkIndices), "delete failed");
					gpnBrowseDrinkIndices = NULL;
				}
//				gnBrowseDrinks = 0;
				gUpdateDrinkList = true;

				// go back to where you came from
				FrmGotoForm(gPreviousForm);

				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == CancelBtn)
			{
				StrArrayFree(gDrinkDescriptors.psArray, gDrinkDescriptors.nItems);
				if(gDrinkDescriptors.psArray)
				{
					ErrFatalDisplayIf(MemPtrFree(gDrinkDescriptors.psArray), "delete failed");
					gDrinkDescriptors.psArray = NULL;
				}
				gDrinkDescriptors.nItems = 0;

				// go back to where you came from
				FrmGotoForm(gPreviousForm);

				isHandled = true;
			}
			break;
	}

	return isHandled;
}

static Boolean AddIngredientFormEventH(const EventPtr event)
{
	FormPtr				pForm = FrmGetActiveForm();
	ListPtr				pList;
	ControlPtr			pControl;
	static char			**psIngredients;
	static char			**psDescriptors;
	static UInt8		nIngredients;
	static UInt8		nDescriptors;
	static UInt8		nAmount;	// in quarter increments 0 to 63 & 3/4
	static char 		sWholeLabel[3],	sFractionalLabel[2];
	static MemHandle	hUnitsList, hUnitsStrings;
	MemPtr				pUnitsList;
	UInt8				n;
	UInt8				*p, *pBlock;
	UInt16				nSize, nLength, nOffset, nCard;
	LocalID				dbID, appInfoID;
	MemHandle			h;
	Boolean				isHandled;
	AppInfoBlockType	info;

	isHandled = false;
	switch(event->eType)
	{
		case frmOpenEvent:
			// get the AppInfo Block, dont forget to free the memory!
			GetAppInfoBlock(&info);

			// fill the ingredient and descriptor arrays
			nIngredients = info.nIngredients;
			psIngredients = (char**)MemPtrNew(nIngredients * sizeof(char*));
			StrPackedCopy(info.pIngredients, psIngredients, nIngredients);
			StrArraySort(&psIngredients, 0, nIngredients-1);

			nDescriptors = info.nIngredientDescriptors;
			psDescriptors = (char**)MemPtrNew((nDescriptors + 1) *
				sizeof(char*));		// add one for "none" selection
			StrPackedCopy(info.pIngredientDescriptors, psDescriptors,
				nDescriptors);
			StrArraySort(&psDescriptors, 0, nDescriptors - 1);
			psDescriptors[nDescriptors] = (char*)MemPtrNew(5 * sizeof(char));
			StrCopy(psDescriptors[nDescriptors], "None");

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			// update the lists
			pList = (ListPtr)GetControlPtr(pForm, IngredientSelectList);
			LstSetListChoices(pList, psIngredients, nIngredients);

			pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
			LstSetListChoices(pList, psDescriptors, nDescriptors + 1);
			LstSetSelection(pList, -1);	// no selection

			// create the units list from the StringTable resource
			//
			hUnitsStrings = DmGetResource('tSTL', UnitsStringTable);
			if(hUnitsStrings)
			{
				pUnitsList = (MemPtr)MemHandleLock(hUnitsStrings);
				n = *(UInt8*)(pUnitsList + StrLen((char*)pUnitsList) + 2);
				hUnitsList = SysFormPointerArrayToStrings(pUnitsList +
					StrLen((char*)pUnitsList) + 3, n);

				pList = (ListPtr)GetControlPtr(pForm, UnitsList);
				LstSetListChoices(pList, MemHandleLock(hUnitsList), n);
				LstSetSelection(pList, noListSelection);
			}

			nAmount = 0;
			pControl = (ControlPtr)GetControlPtr(pForm, WholeAmountBtn);
			CtlSetValue(pControl, 1);

			FrmDrawForm(pForm);
			isHandled = true;
			break;

		case frmCloseEvent:
			AddIngredientCleanup(nIngredients, nDescriptors, psIngredients,
				psDescriptors, hUnitsList, hUnitsStrings);
			break;

		case menuEvent:
			if(event->data.menu.itemID == NewIngredient)
			{
				// get the new ingredient name; if not blank, add it
	  			char sNew[SHORT_STRING];
	  			GetStringModal("New Ingredient", sNew);
	  			if(*sNew == NULL)
	  				break;

	  			// get the AppInfo Block, don't forget to free the memory!
				GetAppInfoBlock(&info);

	  			// test for duplicate ingredient and bail if found
	  			if(StrPackedFind(sNew, info.pIngredients, nIngredients) != -1)
	  			{
	  				DisplayString("The ingredient you entered already exists in the database.");

	  				// free memory associated with the AppInfoBlock
					if(info.pDrinkDescriptors)
						ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
					if(info.pIngredientDescriptors)
						ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
					if(info.pIngredients)
						ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

	  				break;
	  			}

  				// allocate memory for ingredient names & add them to list
				StrArrayFree(psIngredients, nIngredients);
				if(psIngredients)
				{
					ErrFatalDisplayIf(MemPtrFree(psIngredients), "delete failed");
					psIngredients = NULL;
				}
				nIngredients = info.nIngredients;
				psIngredients = StrArrayInsert(sNew, nIngredients,
					info.pIngredients, &nIngredients);

				// allocate space for the new AppInfo Block
				nSize = GetAppInfoBlockSize(&info);
				nSize += StrLen(sNew) + 1;

				h = DmNewHandle(gDrinkDB, nSize);
				ErrFatalDisplayIf(!h, "error creating AppInfo handle");

				appInfoID = MemHandleToLocalID(h);
				DmOpenDatabaseInfo(gDrinkDB, &dbID, 0, 0, &nCard, 0);
				DmSetDatabaseInfo(nCard, dbID, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);
				pBlock = (UInt8*)MemHandleLock(h);

				// write out the new AppInfo Block
				//
				CopyDrinkDescriptors(pBlock, &nOffset, &info);
				CopyIngredientDescriptors(pBlock, &nOffset, &info);

				DmSet(pBlock, nOffset, 1, nIngredients);
				nOffset++;
				for(n = 0; n < nIngredients; n++)
				{
					DmStrCopy(pBlock, nOffset, psIngredients[n]);
					nOffset += StrLen(psIngredients[n]) + 1;
				}
				MemPtrUnlock(pBlock);

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				// sort ingredients alphabetically to display in the list
				StrArraySort(&psIngredients, 0, nIngredients-1);

				pList = (ListPtr)GetControlPtr(pForm, IngredientSelectList);
				LstSetListChoices(pList, psIngredients, nIngredients);
				LstDrawList(pList);			// redraw
  			}

  			if(event->data.menu.itemID == NewDescriptor)
			{
				// get the new descriptor name; if not blank, add it
	  			char sNew[SHORT_STRING];
	  			GetStringModal("New Descriptor", sNew);
	  			if(*sNew == NULL)
	  				break;

	  			// get the AppInfo Block, dont forget to free the memory!
				GetAppInfoBlock(&info);

				// test for duplicate descriptor and bail if found
	  			if(StrPackedFind(sNew, info.pIngredientDescriptors,
	  				info.nIngredientDescriptors) != -1)
	  			{
	  				DisplayString("The descriptor you entered already exists in the database.");

	  				// free memory associated with the AppInfoBlock
					if(info.pDrinkDescriptors)
						ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
					if(info.pIngredientDescriptors)
						ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
					if(info.pIngredients)
						ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

	  				break;
	  			}

				// free the current list of descriptors
  				StrArrayFree(psDescriptors, nDescriptors);
				if(psDescriptors)
				{
					ErrFatalDisplayIf(MemPtrFree(psDescriptors), "delete failed");
					psDescriptors = NULL;
				}
				// allocate memory for descriptor names & add them to the list
				nDescriptors = info.nIngredientDescriptors;
				psDescriptors = StrArrayInsert(sNew, nDescriptors,
					info.pIngredientDescriptors, &nDescriptors);

				// allocate space for the new AppInfo Block
				nSize = GetAppInfoBlockSize(&info);
				nSize += StrLen(sNew) + 1;

				h = DmNewHandle(gDrinkDB, nSize);
				ErrFatalDisplayIf(!h, "Error creating AppInfo handle");

				appInfoID = MemHandleToLocalID(h);
				DmOpenDatabaseInfo(gDrinkDB, &dbID, 0, 0, &nCard, 0);
				DmSetDatabaseInfo(nCard, dbID, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);
				pBlock = (UInt8*)MemHandleLock(h);

				// write out the new AppInfo Block
				//
				CopyDrinkDescriptors(pBlock, &nOffset, &info);

				DmSet(pBlock, nOffset, 1, nDescriptors);
				nOffset++;
				for(n = 0; n < nDescriptors; n++)
				{
					DmStrCopy(pBlock, nOffset, psDescriptors[n]);
					nOffset += StrLen(psDescriptors[n]) + 1;
				}
				CopyIngredients(pBlock, &nOffset, &info);
				MemPtrUnlock(pBlock);

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				// get the newly created appinfo block
				GetAppInfoBlock(&info);

				// rebuild the descriptor list
				StrArrayFree(psDescriptors, nDescriptors);
				if(psDescriptors)
				{
					ErrFatalDisplayIf(MemPtrFree(psDescriptors), "delete failed");
					psDescriptors = NULL;
				}

				nDescriptors = info.nIngredientDescriptors;
				psDescriptors = (char**)MemPtrNew((nDescriptors + 1) *
					sizeof(char*));		// add one for "none" selection
				StrPackedCopy(info.pIngredientDescriptors, psDescriptors,
					nDescriptors);
				StrArraySort(&psDescriptors, 0, nDescriptors - 1);
				psDescriptors[nDescriptors] = (char*)MemPtrNew(5 *
					sizeof(char));
				StrCopy(psDescriptors[nDescriptors], "None");

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
				LstSetListChoices(pList, psDescriptors, nDescriptors + 1);
				LstSetSelection(pList, noListSelection);
	  		}

  			isHandled = true;
			break;

		case keyDownEvent:
			if (event->data.keyDown.chr == vchrPageUp)
			{
				pList = (ListPtr)GetControlPtr(pForm, IngredientSelectList);
				LstScrollList(pList, winUp, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else if (event->data.keyDown.chr == vchrPageDown)
			{
			 	pList = (ListPtr)GetControlPtr(pForm, IngredientSelectList);
				LstScrollList(pList, winDown, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			break;

	  	case ctlRepeatEvent:
			if(event->data.ctlRepeat.controlID == UpScrollBtn)
			{
				pControl = (ControlPtr)GetControlPtr(pForm, WholeAmountBtn);
				if(CtlGetValue(pControl) == 1)
				{
					if(nAmount <= 251)
						nAmount += 4;
					StrIToA(sWholeLabel, nAmount / 4);
					CtlSetLabel(pControl, sWholeLabel);
					CtlDrawControl(pControl);
				}
				else
				{
					pControl = (ControlPtr)GetControlPtr(pForm, FractionalAmountBtn);
					if(nAmount % 4 < 3)
						nAmount++;
					sFractionalLabel[0] = GetFractionalLabel(nAmount);
					sFractionalLabel[1] = NULL;
					CtlSetLabel(pControl, sFractionalLabel);
				}
			}

			if(event->data.ctlRepeat.controlID == DownScrollBtn)
			{
				pControl = (ControlPtr)GetControlPtr(pForm, WholeAmountBtn);
				if(CtlGetValue(pControl) == 1)
				{
					if(nAmount >= 4)
						nAmount -= 4;
					StrIToA(sWholeLabel, nAmount / 4);
					CtlSetLabel(pControl, sWholeLabel);
					CtlDrawControl(pControl);
				}
				else
				{
					pControl = (ControlPtr)GetControlPtr(pForm,
						FractionalAmountBtn);
					if(nAmount % 4 > 0)
						nAmount--;
					sFractionalLabel[0] = GetFractionalLabel(nAmount);
					sFractionalLabel[1] = NULL;
					CtlSetLabel(pControl, sFractionalLabel);
				}
			}
	  		break;

		case ctlSelectEvent:
			if(event->data.ctlEnter.controlID == WholeAmountBtn)
			{
				CtlSetValue((ControlPtr)GetControlPtr(pForm,
					FractionalAmountBtn), 0);
			}
			if(event->data.ctlEnter.controlID == FractionalAmountBtn)
			{
				CtlSetValue((ControlPtr)GetControlPtr(pForm,
					WholeAmountBtn), 0);
			}
			if(event->data.ctlEnter.controlID == DoneBtn)
			{
				char *s;

				if(nAmount == 0)
				{
					DisplayString("Please enter an amount.");
					isHandled = true;
					break;
				}

				GetAppInfoBlock(&info);

				// find the ingredient selection in the AppInfoBlock
				pList = (ListPtr)GetControlPtr(pForm, IngredientSelectList);
				s = LstGetSelectionText(pList, LstGetSelection(pList));
				gIngredient.nName = StrPackedFind(s, info.pIngredients,
					nIngredients);

				gIngredient.nAmount = nAmount;
				pList = (ListPtr)GetControlPtr(pForm, UnitsList);
				if(LstGetSelection(pList) == noListSelection)
					gIngredient.nUnits = 0;
				else
					gIngredient.nUnits = LstGetSelection(pList);

				// get the descriptor selection, if there is one
				pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
				if(LstGetSelection(pList) == noListSelection ||
					LstGetSelection(pList) == nDescriptors)
				{
					gIngredient.nDescriptors = 0;
				}
				else
				{
					gIngredient.nDescriptors = 1;
					s = LstGetSelectionText(pList, LstGetSelection(pList));
					gIngredient.nDescriptor = StrPackedFind(s,
						info.pIngredientDescriptors, nDescriptors + 1);
				}

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				gDrink.nIngredients++;

				// since FrmReturnToForm doesn't generate a frmCloseEvent,
				// lets clean up here before returning
				AddIngredientCleanup(nIngredients, nDescriptors,
					psIngredients, psDescriptors, hUnitsList, hUnitsStrings);
				FrmReturnToForm(AddDrinkForm);

				isHandled = true;
			}
			if(event->data.ctlEnter.controlID == CancelBtn)
			{
				gIngredient.nAmount = 0;

				// since FrmReturnToForm doesn't generate a frmCloseEvent,
				// lets clean up here before returning
				AddIngredientCleanup(nIngredients, nDescriptors,
					psIngredients, psDescriptors, hUnitsList, hUnitsStrings);
				FrmReturnToForm(AddDrinkForm);

				isHandled = true;
			}
			break;
	}

	return isHandled;
}

static Boolean AddDescriptorFormEventH(const EventPtr event)
{
	FormPtr				pForm = FrmGetActiveForm();
	ListPtr				pList;
	void				*pControl;
	static char			sTrigger[SHORT_STRING];
	static char			**psDescriptorOptions;
	static UInt8		nDescriptorOptions;

	UInt8				n;
	UInt8				*p, *pBlock;
	UInt16				nSize, nLength, nOffset, nCard;
	LocalID				dbID, appInfoID;
	MemHandle			h;
	AppInfoBlockType	info;
	Boolean				isHandled;

	isHandled = false;
	switch(event->eType)
	{
	case frmOpenEvent:
		// get the AppInfo Block, dont forget to free the memory!
		GetAppInfoBlock(&info);

		// fill the descriptors array
		if(gDrinkDescriptors.nItems)
		{
			UInt8 nIndex;
			char *s;

			gDrinkDescriptors.psArray = (char**)MemPtrNew(
				gDrinkDescriptors.nItems * sizeof(char*));
			for(n = 0; n < gDrinkDescriptors.nItems; n++)
			{
				s = info.pDrinkDescriptors;
				for(nIndex = 0; nIndex < gDrink.pDescriptors[n]; nIndex++)
					s += StrLen(s) + 1;
				gDrinkDescriptors.psArray[n] = (char*)MemPtrNew(
					(StrLen(s) + 1) * sizeof(char));
				StrCopy(gDrinkDescriptors.psArray[n], s);
			}
			StrArraySort(&gDrinkDescriptors.psArray, 0, gDrinkDescriptors.nItems - 1);
		}
		else
			gDrinkDescriptors.psArray = NULL;

		pList = (ListPtr)GetControlPtr(pForm, DescriptorList);
		LstSetListChoices(pList, gDrinkDescriptors.psArray,	gDrinkDescriptors.nItems);
		LstSetSelection(pList, noListSelection);

		// fill the descriptor options array
		nDescriptorOptions = info.nDrinkDescriptors;
		if(nDescriptorOptions)
		{
			psDescriptorOptions = (char**)MemPtrNew(nDescriptorOptions * sizeof(char*));
			StrPackedCopy(info.pDrinkDescriptors, psDescriptorOptions, nDescriptorOptions);
			StrArraySort(&psDescriptorOptions, 0, nDescriptorOptions - 1);

			// fill in the popup trigger with the first list item
			StrCopy(sTrigger, psDescriptorOptions[0]);
			pControl = GetControlPtr(pForm, DescriptorOptionsPopup);
			CtlSetLabel(pControl, sTrigger);
		}
		else
			psDescriptorOptions = NULL;

		// free memory associated with the AppInfoBlock
		if(info.pDrinkDescriptors)
			ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
		if(info.pIngredientDescriptors)
			ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
		if(info.pIngredients)
			ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

		pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
		LstSetListChoices(pList, psDescriptorOptions, nDescriptorOptions);

		FrmDrawForm(pForm);
		isHandled = true;
		break;

	case frmCloseEvent:
		StrArrayFree(psDescriptorOptions, nDescriptorOptions);
		if(psDescriptorOptions)
		{
			ErrFatalDisplayIf(MemPtrFree(psDescriptorOptions), "delete failed");
			psDescriptorOptions = NULL;
		}
		nDescriptorOptions = 0;
		break;

	case keyDownEvent:
		if (event->data.keyDown.chr == vchrPageUp)
		{
			pList = (ListPtr)GetControlPtr(pForm, DescriptorList);
			LstScrollList(pList, winUp, LstGetVisibleItems(pList) - 1);
		 	isHandled = true;
		}
		else if (event->data.keyDown.chr == vchrPageDown)
		{
		 	pList = (ListPtr)GetControlPtr(pForm, DescriptorList);
			LstScrollList(pList, winDown, LstGetVisibleItems(pList) - 1);
		 	isHandled = true;
		}
		break;

	case ctlSelectEvent:
		if(event->data.ctlEnter.controlID == NewBtn)
		{
			// get the new descriptor name; if not blank, add it
  			char sNew[SHORT_STRING];
  			GetStringModal("New Drink Descriptor", sNew);
  			if(*sNew == NULL)
  				break;

  			// get the AppInfo Block, dont forget to free the memory!
			GetAppInfoBlock(&info);
			nDescriptorOptions = info.nDrinkDescriptors;

			// test for duplicate descriptor and bail if found
  			if(StrPackedFind(sNew, info.pDrinkDescriptors,
  				nDescriptorOptions) != -1)
  			{
  				DisplayString("The descriptor you entered already exists in the database.");

  				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

  				break;
  			}

			// free the current list of descriptors
 			StrArrayFree(psDescriptorOptions, nDescriptorOptions);
			if(psDescriptorOptions)
			{
				ErrFatalDisplayIf(MemPtrFree(psDescriptorOptions), "delete failed");
				psDescriptorOptions = NULL;
			}

			// allocate memory for descriptor names & add them to the list
			psDescriptorOptions = StrArrayInsert(sNew, nDescriptorOptions,
				info.pDrinkDescriptors, &nDescriptorOptions);

			// allocate space for the new AppInfo Block
			nSize = GetAppInfoBlockSize(&info);
			nSize += StrLen(sNew) + 1;
			h = DmNewHandle(gDrinkDB, nSize);
			ErrFatalDisplayIf(!h, "Error creating AppInfo handle");

			appInfoID = MemHandleToLocalID(h);
			DmOpenDatabaseInfo(gDrinkDB, &dbID, 0, 0, &nCard, 0);
			DmSetDatabaseInfo(nCard, dbID, NULL, NULL, NULL, NULL,  NULL,
				NULL, NULL, &appInfoID, NULL, NULL, NULL);
			pBlock = (UInt8*)MemHandleLock(h);

			// write out the new AppInfo Block
			//
			nOffset = 0;
			DmSet(pBlock, nOffset, 1, nDescriptorOptions);
			nOffset++;
			for(n = 0; n < nDescriptorOptions; n++)
			{
				DmStrCopy(pBlock, nOffset, psDescriptorOptions[n]);
				nOffset += StrLen(psDescriptorOptions[n]) + 1;
			}
			CopyIngredientDescriptors(pBlock, &nOffset, &info);
			CopyIngredients(pBlock, &nOffset, &info);
			MemPtrUnlock(pBlock);

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			// get the newly created appinfo block
			GetAppInfoBlock(&info);

			// rebuild the descriptor list
			StrArrayFree(psDescriptorOptions, nDescriptorOptions);
			if(psDescriptorOptions)
			{
				ErrFatalDisplayIf(MemPtrFree(psDescriptorOptions), "delete failed");
				psDescriptorOptions = NULL;
			}

			nDescriptorOptions = info.nDrinkDescriptors;
			psDescriptorOptions = (char**)MemPtrNew(nDescriptorOptions *
				sizeof(char*));
			StrPackedCopy(info.pDrinkDescriptors, psDescriptorOptions,
				nDescriptorOptions);
			StrArraySort(&psDescriptorOptions, 0, nDescriptorOptions - 1);

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			// update the popup item
			CtlSetLabel(GetControlPtr(pForm, DescriptorOptionsPopup),
				psDescriptorOptions[0]);

			// update the list
			pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
			LstSetListChoices(pList, psDescriptorOptions, nDescriptorOptions);
			LstSetSelection(pList, 0);

  			isHandled = true;
  		}
		else if(event->data.ctlEnter.controlID == AddBtn)
		{
			char *s;
			UInt8 *p;
			UInt8 nIndex;
			Int16 nSelect;

			pList = (ListPtr)GetControlPtr(pForm, DescriptorOptionsList);
			nSelect = LstGetSelection(pList);
			if(nSelect > -1)
				s = LstGetSelectionText(pList, nSelect);
			else
				s = LstGetSelectionText(pList, 0);

			// bail if this descriptor has already been added
			for(n = 0; n < gDrinkDescriptors.nItems; n++)
				if(StrCompare(s, gDrinkDescriptors.psArray[n]) == 0)
				{
					DisplayString("This descriptor has already been added. Please select a new descriptor.");
					isHandled = true;
				}
			if(isHandled)
				break;

			// find the index of this descriptor in the appinfo block
			GetAppInfoBlock(&info);
			nSelect = StrPackedFind(s, info.pDrinkDescriptors, info.nDrinkDescriptors);

			// copy the new descriptor sequence into gDrink
			gDrinkDescriptors.nItems++;
			if(gDrink.nDescriptors)
			{
				p = (UInt8*)MemPtrNew(gDrink.nDescriptors * sizeof(UInt8));
				MemMove(p, gDrink.pDescriptors, MemPtrSize(gDrink.pDescriptors));

				ErrFatalDisplayIf(MemPtrFree(gDrink.pDescriptors), "error with free");
				gDrink.nDescriptors = gDrinkDescriptors.nItems;
				gDrink.pDescriptors = (UInt8*)MemPtrNew(gDrink.nDescriptors * sizeof(UInt8));

				MemMove(gDrink.pDescriptors, p, MemPtrSize(p));
				gDrink.pDescriptors[gDrink.nDescriptors - 1] = (UInt8)nSelect;
				ErrFatalDisplayIf(MemPtrFree(p), "error with free");
			}
			else
			{
				gDrink.nDescriptors = 1;
				gDrink.pDescriptors = (UInt8*)MemPtrNew(1 * sizeof(UInt8));
				gDrink.pDescriptors[0] = (UInt8)nSelect;
			}

			// create descriptor list and update ui
			gDrinkDescriptors.psArray = (char**)MemPtrNew(
				gDrinkDescriptors.nItems * sizeof(char*));
			for(n = 0; n < gDrinkDescriptors.nItems; n++)
			{
				s = info.pDrinkDescriptors;
				for(nIndex = 0; nIndex < gDrink.pDescriptors[n]; nIndex++)
					s += StrLen(s) + 1;
				gDrinkDescriptors.psArray[n] = (char*)MemPtrNew(
					(StrLen(s) + 1) * sizeof(char));
				StrCopy(gDrinkDescriptors.psArray[n], s);
			}
			StrArraySort(&gDrinkDescriptors.psArray, 0,
				gDrinkDescriptors.nItems - 1);

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			gDrink.nDescriptors++;

			pList = (ListPtr)GetControlPtr(pForm, DescriptorList);
			LstSetListChoices(pList, gDrinkDescriptors.psArray, gDrinkDescriptors.nItems);
			LstSetSelection(pList, noListSelection);
			FrmDrawForm(pForm);
		}
		else if(event->data.ctlEnter.controlID == RemoveBtn)
		{
			char *s;
			UInt8 *p;
			UInt8 nIndex;
			Int16 nSelect;

			pList = (ListPtr)GetControlPtr(pForm, DescriptorList);

			// bail if there is nothing to remove
			if(LstGetSelection(pList) == noListSelection)
			{
				DisplayString("Please select an item to remove from the descriptors list above.");
				isHandled = true;
				break;
			}

			// find the index of this descriptor in the appinfo block
			GetAppInfoBlock(&info);
			s = LstGetSelectionText(pList, LstGetSelection(pList));
			nSelect = StrPackedFind(s, info.pDrinkDescriptors, nDescriptorOptions);

			// free the current descriptor array
			StrArrayFree(gDrinkDescriptors.psArray, gDrinkDescriptors.nItems);
			if(gDrinkDescriptors.psArray)
			{
				ErrFatalDisplayIf(MemPtrFree(gDrinkDescriptors.psArray), "delete failed");
				gDrinkDescriptors.psArray = NULL;
			}

			// create a new list, excluding the selected descriptor
			gDrinkDescriptors.nItems--;
			nIndex = 0;
			p = (UInt8*)MemPtrNew(gDrinkDescriptors.nItems * sizeof(UInt8));
			for(n = 0; n < gDrinkDescriptors.nItems; n++)
			{
				if(gDrink.pDescriptors[n] != nSelect)
				{
					p[nIndex] = gDrink.pDescriptors[n];
					nIndex++;
				}
			}

			// fill the descriptors array
			if(gDrinkDescriptors.nItems)
			{
				gDrinkDescriptors.psArray = (char**)MemPtrNew(
					gDrinkDescriptors.nItems * sizeof(char*));
				for(n = 0; n < gDrinkDescriptors.nItems; n++)
				{
					s = info.pDrinkDescriptors;
					for(nIndex = 0; nIndex < gDrink.pDescriptors[n]; nIndex++)
						s += StrLen(s) + 1;
					gDrinkDescriptors.psArray[n] = (char*)MemPtrNew(
						(StrLen(s) + 1) * sizeof(char));
					StrCopy(gDrinkDescriptors.psArray[n], s);
				}
				StrArraySort(&gDrinkDescriptors.psArray, 0, gDrinkDescriptors.nItems - 1);
			}
			else
				gDrinkDescriptors.psArray = NULL;

			gDrink.nDescriptors--;

			pList = (ListPtr)GetControlPtr(pForm, DescriptorList);
			LstSetListChoices(pList, gDrinkDescriptors.psArray, gDrinkDescriptors.nItems);
			LstSetSelection(pList, noListSelection);
			FrmDrawForm(pForm);
		}
		else if(event->data.ctlEnter.controlID == DoneBtn)
		{
			// cleanup here since FrmReturnToForm won't generate a frmCloseEvent
			StrArrayFree(psDescriptorOptions, nDescriptorOptions);
			if(psDescriptorOptions)
			{
				ErrFatalDisplayIf(MemPtrFree(psDescriptorOptions), "delete failed");
				psDescriptorOptions = NULL;
			}
			FrmReturnToForm(AddDrinkForm);
			isHandled = true;
		}
		break;
	}

	return isHandled;
}

static Boolean DrinksFormEventH(const EventPtr event)
{
	FormPtr			pForm = FrmGetActiveForm();
	ListPtr			pList;
	void*			pObj;
	UInt16			n, nIndex;
	char			*s, *sRecord;
	MemHandle		hRec;
	Boolean			isHandled = false;
	static Int16	nSelection = 0;
	static Int16	nTopItem = 0;

	switch(event->eType)
	{
		case winEnterEvent:
		case frmOpenEvent:	// the winEnterEvent is not triggered in palmOS5.
			pList = (ListPtr)GetControlPtr(pForm, DrinksList);

			if(gUpdateDrinkList == true)
			{
				DrinkPtr	pDrink;
				UInt16		nIndex, nItems, nDrinksTotal, *pnTemp;
				char		**psTemp;
				Boolean		bInsert;

				StrArrayFree(gpsBrowseDrinks, gnBrowseDrinks);
				if(gpsBrowseDrinks)
				{
					ErrFatalDisplayIf(MemPtrFree(gpsBrowseDrinks), "delete failed");
					gpsBrowseDrinks = NULL;
				}

				if(gpnBrowseDrinkIndices)
				{
					ErrFatalDisplayIf(MemPtrFree(gpnBrowseDrinkIndices), "delete failed");
					gpnBrowseDrinkIndices = NULL;
				}

				nTopItem = 0;
				nSelection = noListSelection;
				LstSetListChoices(pList, NULL, 0);
				FrmDrawForm(pForm);
				DrawBoxedMessage(46, 63, "Please Wait...");

				// create temp memory for a new drinks array
				nDrinksTotal = DmNumRecords(gDrinkDB);
				if(nDrinksTotal)
				{
					pnTemp = (UInt16*)MemPtrNew(nDrinksTotal * sizeof(UInt16));
					psTemp = (char**)MemPtrNew(nDrinksTotal * sizeof(char*));
				}

				// search the database and get the filtered drink list
				nItems = 0;
				for(n = 0; n < nDrinksTotal; n++)
				{
					UInt8 *p;
					MemHandle h;
					IngredientNodePtr pNode;

					// get the next drink record
					pDrink = (DrinkPtr)MemPtrNew(sizeof(DrinkType));
					ErrFatalDisplayIf(GetRecordHandle(&gDrinkDB, n, &h), "couldn't get record");
					p = (UInt8*)MemHandleLock(h);
					UnpackDrinkRecord(p, pDrink);
					MemHandleUnlock(h);

					// if the filters are all off, the drink is automatically added,
					// if any of the filters are on, the filters must be tested
					if(gFilterSettings[0]->action == OFF && gFilterSettings[1]->action == OFF && gFilterSettings[2]->action == OFF)
						bInsert = true;
					else
						bInsert = FilterDrink(gFilterSettings, pDrink);

					// add the drink to the list if appropriate
					if(bInsert == true)
					{
						pnTemp[nItems] = n;
						psTemp[nItems] = (char*)MemPtrNew((StrLen(pDrink->sName) + 1) * sizeof(char));
						StrCopy(psTemp[nItems], pDrink->sName);
						nItems++;
					}
					ErrFatalDisplayIf(DrinkFree(pDrink), "error with free");
					ErrFatalDisplayIf(MemPtrFree(pDrink), "error with free");
				}

				// copy the drink names from the temp storage to psBrowseDrinks
				gnBrowseDrinks = nItems;
				if(gnBrowseDrinks == 0)
				{
					gpnBrowseDrinkIndices = NULL;
					gpsBrowseDrinks = NULL;
				}
				else
				{
					gpnBrowseDrinkIndices = (UInt16*)MemPtrNew(nItems * sizeof(UInt16));
					gpsBrowseDrinks = (char**)MemPtrNew(nItems * sizeof(char*));
					for(n = 0; n < gnBrowseDrinks; n++)
					{
						gpnBrowseDrinkIndices[n] = pnTemp[n];
						gpsBrowseDrinks[n] = (char*)MemPtrNew((StrLen(psTemp[n]) + 1) * sizeof(char));
						StrCopy(gpsBrowseDrinks[n], psTemp[n]);
					}
				}

				// free the temporary memory
				StrArrayFree(psTemp, nItems);
				if(psTemp)
				{
					ErrFatalDisplayIf(MemPtrFree(psTemp), "delete failed");
					psTemp = NULL;
				}
				if(pnTemp)
				{
					ErrFatalDisplayIf(MemPtrFree(pnTemp), "error with free");
					pnTemp = NULL;
				}

				gUpdateDrinkList = false;
			}

			// set the list choices, or if there are none, create the list.
			// using MenuGetActiveMenu() here is a hackish way of preventing the menu
			// from not coming when the list is empty
			if(gpsBrowseDrinks || MenuGetActiveMenu())
			{
				if(gnBrowseDrinks)
					LstSetListChoices(pList, gpsBrowseDrinks, gnBrowseDrinks);
				else
					LstSetListChoices(pList, NULL, 0);
			}


			// draw the form, making sure we don't draw over the menu if visible
			if(MenuGetActiveMenu() == NULL)
			{
				if(gnBrowseDrinks > 0)
				{
					LstSetSelection(pList, nSelection);
					LstSetTopItem(pList, nTopItem);
					FrmDrawForm(pForm);
				}
				else
				{
					FrmDrawForm(pForm);
					DrawBoxedMessage(25, 63, "Filter found no recipes");
    			}
			}
			break;

		case menuEvent:
			if(event->data.menu.itemID == AddDrink)
			{
				gPreviousForm = DrinksForm;
				FrmGotoForm(AddDrinkForm);
			}
			if(event->data.menu.itemID == EditDrink)
			{
				Int16 nSelect;

				pList = (ListPtr)GetControlPtr(pForm, DrinksList);
				nSelect = LstGetSelection(pList);
				if(nSelect == noListSelection)
					DisplayString("Please select a recipe to edit.");
				else
				{
					//
					//
					// add edit handling here
					//
					//
					gPreviousForm = DrinksForm;
					FrmGotoForm(AddDrinkForm);
				}
			}
			if(event->data.menu.itemID == DeleteDrink)
			{
				Int16 nSelect;

				pList = (ListPtr)GetControlPtr(pForm, DrinksList);
				nSelect = LstGetSelection(pList);
				if(nSelect == noListSelection)
				{
					DisplayString("Please select a recipe to delete.");
				}
				else
				{
					if(FrmCustomAlert(AlertQuestion,
						"Are you sure you want to delete this recipe?",	NULL, NULL) == 0)
					{
						DmRemoveRecord(gDrinkDB, gpnBrowseDrinkIndices[nSelect]);

						// free up memory used to store the drink list
						StrArrayFree(gpsBrowseDrinks, gnBrowseDrinks);
						if(gpsBrowseDrinks)
						{
							ErrFatalDisplayIf(MemPtrFree(gpsBrowseDrinks), "delete failed");
							gpsBrowseDrinks = NULL;
						}
						if(gpsBrowseDrinks)
						{
							ErrFatalDisplayIf(MemPtrFree(gpsBrowseDrinks), "delete failed");
							gpsBrowseDrinks = NULL;
						}

						if(gpnBrowseDrinkIndices)
						{
							ErrFatalDisplayIf(MemPtrFree(gpnBrowseDrinkIndices), "delete failed");
						}
						gpnBrowseDrinkIndices = NULL;
						gnBrowseDrinks--;
						gUpdateDrinkList = true;
					}
				}
			}
			break;

		case keyDownEvent:
			if(event->data.keyDown.chr == vchrPageUp)
			{
				pList = (ListPtr)GetControlPtr(pForm, DrinksList);
				LstScrollList(pList, winUp, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else if(event->data.keyDown.chr == vchrPageDown)
			{
			 	pList = (ListPtr)GetControlPtr(pForm, DrinksList);
				LstScrollList(pList, winDown, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else
			{
				char s;
				Boolean isFound = false;

				// scroll the list to the first item starting with of the input
				// character if found
				//
				for(n = 0; n < gnBrowseDrinks; n++)
				{
					s = event->data.keyDown.chr;
					if(StrNCaselessCompare(gpsBrowseDrinks[n], &s, 1) == 0)
					{
						isFound = true;
						break;
					}
				}

				if(isFound)
				{
					pList = (ListPtr)GetControlPtr(pForm, DrinksList);
					LstSetTopItem(pList, n);
					LstDrawList(pList);
				}
			}
			break;

		case lstSelectEvent:
			pList = (ListPtr)GetControlPtr(pForm, DrinksList);
			nSelection = LstGetSelection(pList);
			nTopItem = LstGlueGetTopItem(pList);
			break;

		case ctlSelectEvent:
			if(event->data.ctlEnter.controlID == ViewBtn)
			{
				UInt8 *p;

				pList = (ListPtr)GetControlPtr(pForm, DrinksList);
				nSelection = LstGetSelection(pList);
				if(nSelection == noListSelection)
				{
					isHandled = true;
					break;
				}

				n = gpnBrowseDrinkIndices[nSelection];
				ErrFatalDisplayIf(GetRecordHandle(&gDrinkDB, n, &hRec), "couldn't get record");
				p = (UInt8*)MemHandleLock(hRec);
				UnpackDrinkRecord(p, &gDrink);
				MemHandleUnlock(hRec);

				gPreviousForm = DrinksForm;
				FrmGotoForm(RecipeForm);
		  		isHandled = true;
	  		}
	  		else if(event->data.ctlEnter.controlID == DoneBtn)
			{
				FrmGotoForm(MainForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == FilterBtn)
			{
				FrmPopupForm(FilterForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == InfoBtn)
			{
				FrmHelp(BrowseHelpText);
			}
			break;
	}
	return isHandled;
}

static Boolean FilterFormEventH(const EventPtr event)
{
	FormPtr	pForm = FrmGetActiveForm();
	ListPtr pList;
	UInt8	n;
	char	*s;
	void	*pObj;
	Boolean	isHandled = false;

	n = gnCurrentFilter - 1;
	switch(event->eType)
	{
		case frmOpenEvent:
			// set the title of the filter form
			StrPrintF(gsFilterTitle, "Filter %d of 3", gnCurrentFilter);
			FrmSetTitle(pForm, gsFilterTitle);
			UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
			FrmDrawForm(pForm);
			break;

		case popSelectEvent:
			// remember what is being selected
			if(event->data.popSelect.controlID == FilterDescriptorPopup)
			{
				gFilterSettings[n]->nDescriptorSelection =
					event->data.popSelect.selection;
			}
			break;

		case lstSelectEvent:
			if(gFilterSettings[n]->target == DRINK) {
			    gFilterSettings[n]->nDescriptorSelection =
					event->data.lstEnter.selection;
   			}
   			else if(gFilterSettings[n]->target == INGREDIENT) {
				gFilterSettings[n]->nIngredientSelection =
					event->data.lstEnter.selection;
      		}
			break;

		case ctlRepeatEvent:
			if(event->data.ctlRepeat.controlID == LeftScrollBtn && gnCurrentFilter > 1)
			{
				gnCurrentFilter--;

				// set the title of the filter form
				StrPrintF(gsFilterTitle, "Filter %d of 3", gnCurrentFilter);
				FrmSetTitle(pForm, gsFilterTitle);

				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			if(event->data.ctlRepeat.controlID == RightScrollBtn && gnCurrentFilter < nFILTERS)
			{
				gnCurrentFilter++;

				// set the title of the filter form
				StrPrintF(gsFilterTitle, "Filter %d of 3", gnCurrentFilter);
				FrmSetTitle(pForm, gsFilterTitle);

				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			break;

		case ctlSelectEvent:
			if(event->data.ctlEnter.controlID == FilterActionOffPB &&
				gFilterSettings[n]->action != OFF)
			{
				gFilterSettings[n]->action = OFF;
				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == FilterDrinkPB &&
				gFilterSettings[n]->target != DRINK)
			{
				gFilterSettings[n]->target = DRINK;

				// set the default descriptor to the first in the list
				gFilterSettings[n]->nDescriptorSelection = 0;

				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == FilterIngredientPB &&
				gFilterSettings[n]->target != INGREDIENT)
			{
				AppInfoBlockType info;

				// set the default descriptor to the last in list (None)
				GetAppInfoBlock(&info);
				gFilterSettings[n]->nDescriptorSelection = info.nIngredientDescriptors;

				// free memory associated with the AppInfoBlock
				if(info.pDrinkDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
				if(info.pIngredientDescriptors)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
				if(info.pIngredients)
					ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

				gFilterSettings[n]->target = INGREDIENT;
				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == FilterActionInclude)
			{
				gFilterSettings[n]->action = INCLUDE;
				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == FilterActionExclude)
			{
				gFilterSettings[n]->action = EXCLUDE;
				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == FilterTypeCheckbox) {
				gFilterSettings[n]->bTypeChecked = !gFilterSettings[n]->bTypeChecked;
				UpdateFilterUI(pForm, gFilterSettings, gnCurrentFilter);
				FrmDrawForm(pForm);
			}
			else if(event->data.ctlEnter.controlID == DoneBtn)
			{
				UInt16 n;

				gnCurrentFilter = 1;
				gUpdateDrinkList = true;
				FrmReturnToForm(DrinksForm);
				isHandled = true;
			}
			else if(event->data.ctlEnter.controlID == CancelBtn)
			{
				gnCurrentFilter = 1;
				FrmReturnToForm(DrinksForm);
				isHandled = true;
			}
			break;

		case keyDownEvent:
			if(event->data.keyDown.chr == vchrPageUp)
			{
				pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
				LstScrollList(pList, winUp, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else if(event->data.keyDown.chr == vchrPageDown)
			{
			 	pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
				LstScrollList(pList, winDown, LstGetVisibleItems(pList) - 1);
			 	isHandled = true;
			}
			else
			{
				char s;
				UInt16 nIndex = 0;
				Boolean isFound = false;

				if(gFilterSettings[n]->target == INGREDIENT)
				{
					// scroll the list to the first item starting with of the input
					// character if found
					for(nIndex = 0; nIndex < gFilterSettings[n]->nIngredients; nIndex++)
					{
						s = event->data.keyDown.chr;
						if(StrNCaselessCompare(gFilterSettings[n]->psIngredients[nIndex], &s, 1) == 0)
						{
							isFound = true;
							break;
						}
					}

					if(isFound)
					{
						pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
						LstSetTopItem(pList, nIndex);
						LstSetSelection(pList, nIndex);
						gFilterSettings[n]->nIngredientSelection = nIndex;
						LstDrawList(pList);
					}
				}
				else if(gFilterSettings[n]->target == DRINK)
				{
					// scroll the list to the first item starting with of the input
					// character if found
					for(nIndex = 0; nIndex < gFilterSettings[n]->nDescriptors; nIndex++)
					{
						s = event->data.keyDown.chr;
						if(StrNCaselessCompare(gFilterSettings[n]->psDescriptors[nIndex], &s, 1) == 0)
						{
							isFound = true;
							break;
						}
					}

					if(isFound)
					{
						pList = (ListPtr)GetControlPtr(pForm, FilterIngredientList);
						LstSetTopItem(pList, nIndex);
						LstSetSelection(pList, nIndex);
						gFilterSettings[n]->nDescriptorSelection = nIndex;
						LstDrawList(pList);
					}
    			}

				isHandled = true;
			}
			break;
	}

	return isHandled;
}

static Boolean RecipeFormEventH(const EventPtr event)
{
	FormPtr				pForm = FrmGetActiveForm();
	ScrollBarPtr		pScroll = (ScrollBarPtr)GetControlPtr(pForm, RecipeScroll);
	FieldPtr			pField = (FieldPtr)GetControlPtr(pForm, RecipeField);
	UInt16				n, nSearch, nIndex;
	char 				sNew[MEDIUM_STRING];
	static char			s[LONG_STRING];
	IngredientNodePtr	pIngredient;
	AppInfoBlockType	info;
	Boolean				isHandled = false;

	switch(event->eType)
	{
		case frmOpenEvent:
			// build the output string and display it in the text field
			FrmSetTitle(pForm, gDrink.sName);
			StrCopy(s, "Ingredients:\n");
			StrCat(s, "\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\n");

			GetAppInfoBlock(&info);
			pIngredient = gDrink.pIngredientList;
			for(n = 0; n < gDrink.nIngredients; n++)
			{
				BuildIngredientString(&info, pIngredient->ingredient, sNew);
				StrCat(s, "\x95 ");
				StrCat(s, sNew);
				StrCat(s, "\n");
				if(pIngredient->pNext)
					pIngredient = pIngredient->pNext;
			}

			// free memory associated with the AppInfoBlock
			if(info.pDrinkDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pDrinkDescriptors), "error deleting");
			if(info.pIngredientDescriptors)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredientDescriptors), "error deleting");
			if(info.pIngredients)
				ErrFatalDisplayIf(MemPtrFree(info.pIngredients), "error deleting");

			StrCat(s, "\nDirections:\n");
			StrCat(s, "\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\x97\n");
			StrCat(s, gDrink.sDirections);
			SetFieldText(pForm, RecipeField, s);
			UpdateScrollbar(pField, pScroll);

			FrmDrawForm(pForm);
			isHandled = true;
			break;

		case frmCloseEvent:
			// clean up
			ErrFatalDisplayIf(DrinkFree(&gDrink), "error freeing drink");
			break;

		case keyDownEvent:
			if(event->data.keyDown.chr == vchrPageUp && FldScrollable(pField, winUp))
			{
				ScrollLines(pField, pScroll, -1, true);
				UpdateScrollbar(pField, pScroll);
			 	isHandled = true;
			}
			else if(event->data.keyDown.chr == vchrPageDown && FldScrollable(pField, winDown))
			{
			 	ScrollLines(pField, pScroll, 1, true);
			 	UpdateScrollbar(pField, pScroll);
			 	isHandled = true;
			}
			break;

		case ctlSelectEvent:
			if (event->data.ctlEnter.controlID == DoneBtn)
			{
				// go back to where you came from
				FrmGotoForm(gPreviousForm);
				isHandled = true;
			}
			break;

		case sclRepeatEvent:
			ScrollLines(pField, pScroll, event->data.sclRepeat.newValue - event->data.sclRepeat.value, false);
			break;
	}

	return isHandled;
}

//===========================================================================
//	System Boilerplate...
//===========================================================================

static Err StartApplication(void)
{
	LocalID		lid;
	Int16		cCards;
	int 		n;
	char*		sRecord;
	MemHandle	h;
	Err			error;

	// set up the screen to handle the best color depth
	SetScreenMode();

	// create database, if it hasn't been created already
	h = DmGetResource('dbas', 1);
	if (h != NULL)
	{
		void *pv = MemHandleLock(h);
		if(pv != NULL)
			DmCreateDatabaseFromImage(pv);
		MemHandleUnlock(h);
		DmReleaseResource(h);
	}

	// get a reference to the database
	lid = 0;
	cCards = MemNumCards();
	for(n = 0; n < cCards; n++)
	{
		lid = DmFindDatabase(n, "DrinkDB-Minibar");
		if(lid != 0)
      		break;
	}
	ErrFatalDisplayIf(lid == 0, "couldn't find DrinkDB database");
	gDrinkDB = DmOpenDatabase(n, lid, dmModeReadWrite);
	ErrFatalDisplayIf(gDrinkDB == NULL, "couldn't open DrinkDB database");

	// initialize other global variables
	//
	gIngredient.nAmount = 0;
	gPreviousForm = MainForm;

	gFilterSettings = (FilterType**)MemPtrNew(nFILTERS * sizeof(FilterPtr));
	for(n = 0; n < nFILTERS; n++)
	{
		gFilterSettings[n] = (FilterPtr)MemPtrNew(sizeof(FilterType));
		gFilterSettings[n]->action = OFF;
		gFilterSettings[n]->target = DRINK;
		gFilterSettings[n]->nDescriptors = 0;
		gFilterSettings[n]->nIngredients = 0;
		gFilterSettings[n]->nDescriptorSelection = 0;
		gFilterSettings[n]->nIngredientSelection = 0;
		gFilterSettings[n]->bTypeChecked = false;
		gFilterSettings[n]->psDescriptors = NULL;
		gFilterSettings[n]->psIngredients = NULL;
	}

	gnBrowseDrinks = 0;
	gpsBrowseDrinks = NULL;
	gpnBrowseDrinkIndices = NULL;
	gnCurrentFilter =  1;

	gUpdateDrinkList = true;

	gsNone = (char*)MemPtrNew(6 * sizeof(char));
	StrCopy(gsNone, "None");

	gsFilterTitle = (char*)MemPtrNew(14 * sizeof(char));
	StrCopy(gsFilterTitle, "Filter 1 of 3");

	SysRandom(TimGetTicks());
	FrmGotoForm(MainForm);

	return 0;
}

static void StopApplication(void)
{
	UInt16 n;

	// free up memory used to store the drink list
	StrArrayFree(gpsBrowseDrinks, gnBrowseDrinks);
	if(gpsBrowseDrinks)
	{
		ErrFatalDisplayIf(MemPtrFree(gpsBrowseDrinks), "delete failed");
		gpsBrowseDrinks = NULL;
	}

	if(gpnBrowseDrinkIndices)
	{
		ErrFatalDisplayIf(MemPtrFree(gpnBrowseDrinkIndices), "delete failed");
	}
	gpnBrowseDrinkIndices = NULL;

	// free memory associated with the filters
	for(n = 0; n < nFILTERS; n++)
	{
		StrArrayFree(gFilterSettings[n]->psDescriptors, gFilterSettings[n]->nDescriptors);
		if(gFilterSettings[n]->psDescriptors)
		{
			ErrFatalDisplayIf(MemPtrFree(gFilterSettings[n]->psDescriptors), "delete failed");
			gFilterSettings[n]->psDescriptors = NULL;
		}

		StrArrayFree(gFilterSettings[n]->psIngredients, gFilterSettings[n]->nIngredients);
		if(gFilterSettings[n]->psIngredients)
		{
			ErrFatalDisplayIf(MemPtrFree(gFilterSettings[n]->psIngredients), "delete failed");
			gFilterSettings[n]->psIngredients = NULL;
		}

		if(gFilterSettings[n])
			ErrFatalDisplayIf(MemPtrFree(gFilterSettings[n]), "error with free");
	}
	if(gFilterSettings)
		ErrFatalDisplayIf(MemPtrFree(gFilterSettings), "error with free");

	ErrFatalDisplayIf(MemPtrFree(gsNone), "error with free");
	ErrFatalDisplayIf(MemPtrFree(gsFilterTitle), "error with free");
	ErrFatalDisplayIf(DmCloseDatabase(gDrinkDB), "error closing database on exit");
	FrmCloseAllForms();
}

static void EventLoop(void)
{
	EventType	event;
	short		error;
	FormPtr		pForm;
	Int16		formID;

	do
	{
		EvtGetEvent(&event, SysTicksPerSecond() / 2);
		if(SysHandleEvent(&event))
			continue;

		if(MenuHandleEvent((void *)0, &event, &error))
			continue;

		if(event.eType == frmLoadEvent)
		{
			formID = event.data.frmLoad.formID;
			pForm = FrmInitForm(formID);
			FrmSetActiveForm(pForm);

			switch(formID)
			{
				case MainForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)MainFormEventH);
					break;
				case AddDrinkForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)AddDrinkFormEventH);
					break;
				case AddIngredientForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)AddIngredientFormEventH);
					break;
				case AddDescriptorForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)AddDescriptorFormEventH);
					break;
				case DrinksForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)DrinksFormEventH);
					break;
				case FilterForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)FilterFormEventH);
					break;
				case RecipeForm:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr)RecipeFormEventH);
					break;
			}
		}
		FrmDispatchEvent(&event);
	}
	while (event.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	Int16 error = 0;

	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		error = StartApplication();		// Application start code
		if (error)
			return error;
		EventLoop();
		StopApplication();				// Application stop code
	}
	return 0;
}

// EOF
